import logging
import re
from pathlib import Path

from cg.constants.constants import FileExtensions, ReadDirection
from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.files.abstract import FileFormatter
from cg.services.deliver_files.file_formatter.files.models import FastqFile
from cg.services.deliver_files.file_formatter.files.sample_service import FileManager
from cg.services.deliver_files.file_formatter.path_name.abstract import PathNameFormatter
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.fastq_concatenation_service.utils import generate_concatenated_fastq_delivery_path
from cg.utils.files import get_all_files_in_directory_tree

LOG = logging.getLogger(__name__)


class SampleFileConcatenationFormatter(FileFormatter):
    """
    Format the sample files to deliver, concatenate fastq files and return the formatted files.
    Used for workflows: Microsalt.
    args:
        file_manager: The file manager
        path_name_formatter: The path name formatter to format paths to either a flat or nested structure in the delivery destination
        concatenation_service: The fastq concatenation service to concatenate fastq files.
    """

    def __init__(
        self,
        file_manager: FileManager,
        path_name_formatter: PathNameFormatter,
        concatenation_service: FastqConcatenationService,
    ):
        self.file_manager = file_manager
        self.path_name_formatter = path_name_formatter
        self.concatenation_service = concatenation_service

    def format_files(
        self, moved_files: list[SampleFile], delivery_path: Path
    ) -> list[FormattedFile]:
        """
        Format the sample files to deliver, concatenate fastq files and return the formatted files.
        args:
            moved_files: list[SampleFile]: List of sample files to deliver.
            These are files that have been moved from housekeeper to the delivery path.
            delivery_path: Path: Path to the delivery directory.
        """
        LOG.debug("[FORMAT SERVICE] Formatting and concatenating sample files")
        sample_names: set[str] = self._get_sample_names(sample_files=moved_files)
        self._create_sample_directories(delivery_path=delivery_path, sample_names=sample_names)
        formatted_files: list[FormattedFile] = self._format_sample_file_paths(moved_files)
        LOG.debug(
            f"[FORMAT SERVICE] number of formatted files: {len(formatted_files)}, number of moved files: {len(moved_files)}"
        )
        self._rename_original_files(formatted_files)
        LOG.debug(f"[FORMAT SERVICE] delivery_path: {delivery_path}")
        concatenation_map: dict[Path, Path] = self._concatenate_fastq_files(
            delivery_path=delivery_path,
            sample_names=sample_names,
        )
        self._replace_fastq_paths(
            concatenation_maps=concatenation_map,
            formatted_files=formatted_files,
        )
        return formatted_files

    @staticmethod
    def _get_sample_names(sample_files: list[SampleFile]) -> set[str]:
        """Extract sample names from the sample files."""
        return {sample_file.sample_name for sample_file in sample_files}

    def _create_sample_directories(self, sample_names: set[str], delivery_path: Path) -> None:
        """Create directories for each sample name only if the file name formatter is the NestedSampleFileFormatter.
        args:
            sample_names: set[str]: Set of sample names.
            delivery_path: Path: Path to the delivery directory.
        """
        if not isinstance(self.path_name_formatter, NestedStructurePathFormatter):
            return
        for sample_name in sample_names:
            self.file_manager.create_directories(base_path=delivery_path, directories={sample_name})

    def _format_sample_file_paths(self, sample_files: list[SampleFile]) -> list[FormattedFile]:
        """
        Return a list of formatted sample files.
        args:
            sample_files: The sample files to format
        """
        return [
            FormattedFile(
                original_path=sample_file.file_path,
                formatted_path=self.path_name_formatter.format_file_path(
                    file_path=sample_file.file_path,
                    provided_id=sample_file.sample_id,
                    provided_name=sample_file.sample_name,
                ),
            )
            for sample_file in sample_files
        ]

    def _rename_original_files(self, formatted_files: list[FormattedFile]) -> None:
        """
        Rename the formatted files.
        args:
            formatted_files: list[FormattedFile]: List of formatted files.
        """
        LOG.debug("[FORMAT SERVICE] Renaming original files")
        for formatted_file in formatted_files:
            self.file_manager.rename_file(
                src=formatted_file.original_path, dst=formatted_file.formatted_path
            )

    def _concatenate_fastq_files(
        self, delivery_path: Path, sample_names: set[str]
    ) -> dict[Path, Path]:
        """Concatenate fastq files for each sample and return the forward and reverse concatenated paths.
        args:
            delivery_path: Path: Path to the delivery directory.
            sample_names: set[str]: Set of sample names.
        returns:
            dict[Path, Path]: Dictionary with the original fastq file path as key and the concatenated path as value.
        """
        LOG.debug(f"[FORMAT SERVICE] delivery_path: {delivery_path}")
        fastq_files: list[FastqFile] = self._get_unique_sample_fastq_paths(
            sample_names=sample_names, delivery_path=delivery_path
        )
        grouped_fastq_files: dict[str, list[FastqFile]] = self._group_fastq_files_per_sample(
            sample_names=sample_names, fastq_files=fastq_files
        )
        concatenation_maps: dict[Path, Path] = {}
        for sample in grouped_fastq_files.keys():
            fastq_directory: Path = grouped_fastq_files[sample][0].fastq_file_path.parent
            forward_path: Path = generate_concatenated_fastq_delivery_path(
                fastq_directory=fastq_directory,
                sample_name=sample,
                direction=ReadDirection.FORWARD,
            )
            reverse_path: Path = generate_concatenated_fastq_delivery_path(
                fastq_directory=fastq_directory,
                sample_name=sample,
                direction=ReadDirection.REVERSE,
            )
            self.concatenation_service.concatenate(
                sample_id=sample,
                fastq_directory=fastq_directory,
                forward_output_path=forward_path,
                reverse_output_path=reverse_path,
                remove_raw=True,
            )
            concatenation_maps.update(
                self._get_concatenation_map(
                    forward_path=forward_path,
                    reverse_path=reverse_path,
                    fastq_files=grouped_fastq_files[sample],
                )
            )
        return concatenation_maps

    def _get_unique_sample_fastq_paths(
        self, sample_names: set[str], delivery_path: Path
    ) -> list[FastqFile]:
        """
        Get a list of unique sample fastq file paths given a delivery path.
        args:
            sample_names: set[str]: Set of sample names.
            delivery_path: Path: Path to the delivery directory
        returns:
            list[FastqFile]: List of FastqFile objects.
        """
        sample_paths: list[FastqFile] = []
        LOG.debug(
            f"[CONCATENATION SERVICE] Getting unique sample fastq file paths in {delivery_path}"
        )
        list_of_files: list[Path] = get_all_files_in_directory_tree(delivery_path)
        for sample_name in sample_names:
            for file in list_of_files:
                if self._has_expected_sample_name_format_match(
                    sample_name=sample_name, file_path=file
                ) and self._is_lane_fastq_file(file):
                    LOG.debug(
                        f"[CONCATENATION SERVICE] Found fastq file: {file} for sample: {sample_name}"
                    )
                    sample_paths.append(
                        FastqFile(
                            fastq_file_path=Path(delivery_path, file),
                            sample_name=sample_name,
                            read_direction=self._determine_read_direction(file),
                        )
                    )
        if not sample_paths:
            raise FileNotFoundError(
                f"Could not find any fastq files to concatenate in {delivery_path}."
            )
        return sample_paths

    @staticmethod
    def _has_expected_sample_name_format_match(sample_name: str, file_path: Path) -> bool:
        """
        Check if the sample name is an exact match in the file path.
        Fastq files are expected to have the sample name in the file path formatted as such: _{sample_name}_
        args:
            sample_name: str: The sample name to match.
            file_path: Path: The file path to check.
        """
        return f"_{sample_name}_" in file_path.as_posix()

    @staticmethod
    def _get_concatenation_map(
        forward_path: Path, reverse_path: Path, fastq_files: list[FastqFile]
    ) -> dict[Path, Path]:
        """
        Get a list of ConcatenationMap objects for a sample.
        NOTE: the fastq_files must be grouped by sample name.
        args:
            forward_path: Path: Path to the forward concatenated file.
            reverse_path: Path: Path to the reverse concatenated file.
            fastq_files: list[FastqFile]: List of fastq files for a single ample.
        """
        concatenation_map: dict[Path, Path] = {}
        for fastq_file in fastq_files:
            concatenation_map[fastq_file.fastq_file_path] = (
                forward_path if fastq_file.read_direction == ReadDirection.FORWARD else reverse_path
            )
        return concatenation_map

    @staticmethod
    def _determine_read_direction(fastq_path: Path) -> ReadDirection:
        """Determine the read direction of a fastq file.
            Assumes that the fastq file path contains 'R1' or 'R2' to determine the read direction.
        args:
            fastq_path: Path: Path to the fastq file.
        """
        if f"R{ReadDirection.FORWARD}" in fastq_path.as_posix():
            return ReadDirection.FORWARD
        return ReadDirection.REVERSE

    def _group_fastq_files_per_sample(
        self, sample_names: set[str], fastq_files: list[FastqFile]
    ) -> dict[str, list[FastqFile]]:
        """Group fastq files per sample.
        returns a dictionary with sample names as keys and a list of fastq files as values.
        args:
            sample_names: set[str]: Set of sample names.
            fastq_files: list[FastqFile]: List of fastq files.
        """

        sample_fastq_files: dict[str, list[FastqFile]] = {
            sample_name: [] for sample_name in sample_names
        }
        for fastq_file in fastq_files:
            sample_fastq_files[fastq_file.sample_name].append(fastq_file)
        self._validate_sample_fastq_file_share_same_directory(sample_fastq_files=sample_fastq_files)
        return sample_fastq_files

    def _replace_fastq_paths(
        self,
        concatenation_maps: dict[Path, Path],
        formatted_files: list[FormattedFile],
    ) -> None:
        """
        Replace the fastq file paths with the new concatenated fastq file paths.
        Uses the concatenation map with the formatted file path as key and the concatenated path as value.
        args:
            concatenation_maps: list[ConcatenationMap]: List of ConcatenationMap objects.
            formatted_files: list[FormattedFile]: List of formatted files.
        """
        for formatted_file in formatted_files:
            if self._is_lane_fastq_file(formatted_file.formatted_path):
                formatted_file.formatted_path = concatenation_maps[formatted_file.formatted_path]

    @staticmethod
    def _validate_sample_fastq_file_share_same_directory(
        sample_fastq_files: dict[str, list[FastqFile]],
    ) -> None:
        """
        Assert that all fastq files for a sample share the same directory.
        This is to ensure that the files are concatenated within the expected directory path.
        raises: ValueError if the fastq files are not in the same directory.
        args:
            sample_fastq_files: dict[str, list[FastqFile]]: Dictionary of sample names and their fastq files.
        """
        for sample_name in sample_fastq_files.keys():
            fastq_files: list[FastqFile] = sample_fastq_files[sample_name]
            parent_dir: Path = fastq_files[0].fastq_file_path.parent
            for fastq_file in fastq_files:
                if fastq_file.fastq_file_path.parent != parent_dir:
                    raise ValueError(
                        f"Sample {sample_name} fastq files are not in the same directory. "
                        f"Cannot concatenate. It will would result in sporadic file paths."
                    )

    @staticmethod
    def _is_lane_fastq_file(file_path: Path) -> bool:
        """Check if a fastq file is a from a lane and read direction.
        Note pattern: *_L[0-9]{3}_R[1-2]_[0-9]{3}.fastq.gz
        *_ is a wildcard for the flow cell id followed by sample name.
        L[0-9]{3} is the lane number, i.e. L001, L002 etc.
        R[1-2] is the read direction, i.e. R1 or R2.
        [0-9]{3} is the trailing three digits after read direction.
        args:
            file_path: Path: Path to the fastq file.
        """

        pattern = f".*_L[0-9]{{3}}_R[1-2]_[0-9]{{3}}{FileExtensions.FASTQ}{FileExtensions.GZIP}"
        return (
            re.fullmatch(
                pattern=pattern,
                string=file_path.name,
            )
            is not None
        )



"""API for encryption on Hasta"""

import logging
from pathlib import Path

from cg.constants import FileExtensions
from cg.constants.encryption import EncryptionUserID, GPGParameters
from cg.utils import Process

LOG = logging.getLogger(__name__)
LIMIT_PIGZ_TASK: int = 3


class EncryptionAPI:
    """Class that uses gpg for various encryption and decryption functionality"""

    def __init__(self, binary_path: str, dry_run: bool = False):
        self.binary_path: str = binary_path
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    def run_gpg_command(self, command: list) -> None:
        """Runs a GPG command"""
        LOG.info("Starting GPG command:")
        LOG.info(f"{self.process.binary} {' '.join(command)}")
        self.process.run_command(command, dry_run=self.dry_run)

    def get_symmetric_passphrase_cmd(
        self, passphrase_file_path: Path, quality_level: int = 2, count: int = 256
    ) -> str:
        """Return command to generate a symmetrical passphrase file."""
        return " ".join(
            [
                self.binary_path,
                "--gen-random",
                str(quality_level),
                str(count),
                ">",
                passphrase_file_path.as_posix(),
            ]
        )

    def get_asymmetrically_encrypt_passphrase_cmd(self, passphrase_file_path: Path) -> str:
        """Return command to asymmetrically encrypt a symmetrical passphrase file."""
        return " ".join(
            [
                self.binary_path,
                "--encrypt",
                "--yes",
                "--recipient",
                EncryptionUserID.HASTA_USER_ID,
                "--output",
                passphrase_file_path.with_suffix(FileExtensions.GPG).as_posix(),
                passphrase_file_path.as_posix(),
            ]
        )

    def get_asymmetric_decryption_command(self, input_file: Path, output_file: Path) -> list[str]:
        """Generates the gpg command for asymmetric decryption"""
        decryption_parameters: list = GPGParameters.ASYMMETRIC_DECRYPTION.copy()
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        decryption_parameters.extend(output_parameter)
        return decryption_parameters

    def get_symmetric_decryption_command(
        self, input_file: Path, output_file: Path, encryption_key: Path
    ) -> list[str]:
        """Generates the gpg command for symmetric decryption"""
        decryption_parameters: list = GPGParameters.SYMMETRIC_DECRYPTION.copy()
        decryption_parameters.append(str(encryption_key))
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        decryption_parameters.extend(output_parameter)
        return decryption_parameters

    def create_pending_file(self, pending_path: Path) -> None:
        """Create a pending flag file."""
        LOG.info(f"Creating pending flag {pending_path}")
        if self.dry_run:
            return
        pending_path.touch(exist_ok=False)



from cg.models.orders.constants import OrderType
from cg.models.orders.excel_sample import ExcelSample

ORDER_TYPES_WITH_CASES = [
    OrderType.BALSAMIC,
    OrderType.BALSAMIC_UMI,
    OrderType.MIP_DNA,
    OrderType.MIP_RNA,
    OrderType.NALLO,
    OrderType.RAREDISEASE,
    OrderType.RNAFUSION,
    OrderType.TOMTE,
]


def are_all_samples_metagenome(samples: list[ExcelSample]) -> bool:
    """Check if all samples are metagenome samples"""
    return all(sample.application.startswith("ME") for sample in samples)



"""Module for the arnoldAPIClient."""

import logging
from http import HTTPStatus

import requests
from requests import Response

from cg.clients.arnold.dto.create_case_request import CreateCaseRequest
from cg.clients.arnold.exceptions import ArnoldClientError, ArnoldServerError

LOG = logging.getLogger(__name__)


class ArnoldAPIClient:
    def __init__(self, config: dict):
        self.api_url: str = config["arnold"]["api_url"]

    def create_case(self, case: CreateCaseRequest) -> None:
        endpoint: str = f"{self.api_url}/case/"
        post_request_data: CreateCaseRequest = case
        response: Response = requests.post(
            endpoint, data=post_request_data.model_dump_json(), verify=True
        )
        if response.status_code == HTTPStatus.OK:
            LOG.info(f"Info {response.content}.")
            return
        self._handle_errors(response)

    @staticmethod
    def _handle_errors(response: Response):
        if HTTPStatus.BAD_REQUEST <= response.status_code < HTTPStatus.INTERNAL_SERVER_ERROR:
            LOG.error(f"Client error: {response.status_code}. Reason {response.content}")
            raise ArnoldClientError

        elif HTTPStatus.INTERNAL_SERVER_ERROR <= response.status_code:
            LOG.error(f"Server error: {response.status_code}. Reason {response.content}")
            raise ArnoldServerError



import logging
import subprocess
from pathlib import Path
from subprocess import CalledProcessError

from pydantic import EmailStr

from cg.apps.lims.api import LimsAPI
from cg.constants import SexOptions
from cg.constants.constants import BedVersionGenomeVersion, GenomeVersion
from cg.constants.process import EXIT_SUCCESS
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.models.cg_config import BalsamicConfig
from cg.services.analysis_starter.configurator.models.balsamic import (
    BalsamicConfigInput,
    BalsamicConfigInputPanel,
    BalsamicConfigInputWGS,
)
from cg.store.models import BedVersion, Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class BalsamicConfigFileCreator:

    def __init__(self, cg_balsamic_config: BalsamicConfig, lims_api: LimsAPI, status_db: Store):
        self.status_db = status_db
        self.root_dir = cg_balsamic_config.root
        self.lims_api: LimsAPI = lims_api
        self.conda_binary: Path = cg_balsamic_config.conda_binary
        self.conda_env: str = cg_balsamic_config.conda_env
        self.balsamic_binary: Path = cg_balsamic_config.binary_path
        self.root_dir: Path = cg_balsamic_config.root
        self.bed_directory: Path = cg_balsamic_config.bed_path
        self.cache_dir: Path = cg_balsamic_config.balsamic_cache
        self.cache_version: str = cg_balsamic_config.cache_version
        self.cadd_path: Path = cg_balsamic_config.cadd_path
        self.genome_interval_path: Path = cg_balsamic_config.genome_interval_path
        self.gens_coverage_female_path: Path = cg_balsamic_config.gens_coverage_female_path
        self.gens_coverage_male_path: Path = cg_balsamic_config.gens_coverage_male_path
        self.gnomad_af5_path: Path = cg_balsamic_config.gnomad_af5_path
        self.environment: str = cg_balsamic_config.conda_env
        self.sentieon_licence_path: Path = cg_balsamic_config.sentieon_licence_path
        self.sentieon_licence_server: str = cg_balsamic_config.sentieon_licence_server
        self.loqusdb_artefact_snv: Path = cg_balsamic_config.loqusdb_dump_files.artefact_snv
        self.artefact_sv_observations: Path = cg_balsamic_config.loqusdb_dump_files.artefact_sv
        self.loqusdb_cancer_germline_snv: Path = (
            cg_balsamic_config.loqusdb_dump_files.cancer_germline_snv
        )
        self.loqusdb_cancer_somatic_snv: Path = (
            cg_balsamic_config.loqusdb_dump_files.cancer_somatic_snv
        )
        self.loqusdb_cancer_somatic_snv_panels: dict = (
            cg_balsamic_config.loqusdb_dump_files.cancer_somatic_snv_panels
        )
        self.loqusdb_cancer_somatic_sv: Path = (
            cg_balsamic_config.loqusdb_dump_files.cancer_somatic_sv
        )
        self.loqusdb_clinical_snv: Path = cg_balsamic_config.loqusdb_dump_files.clinical_snv
        self.loqusdb_clinical_sv: Path = cg_balsamic_config.loqusdb_dump_files.clinical_sv
        self.panel_of_normals: dict = cg_balsamic_config.panel_of_normals
        self.slurm_account: str = cg_balsamic_config.slurm.account
        self.slurm_mail_user: EmailStr = cg_balsamic_config.slurm.mail_user
        self.swegen_snv: Path = cg_balsamic_config.swegen_snv
        self.swegen_sv: Path = cg_balsamic_config.swegen_sv

    def create(self, case_id: str, fastq_path: Path, **flags) -> None:
        config_cli_input: BalsamicConfigInput = self._build_config_input(
            case_id=case_id, fastq_path=fastq_path, **flags
        )
        self._create_config_file(config_cli_input)

    @staticmethod
    def _create_config_file(config_cli_input: BalsamicConfigInput) -> None:
        final_command: str = config_cli_input.dump_to_cli()
        LOG.debug(f"Running: {final_command}")
        result = subprocess.run(
            args=final_command,
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if result.returncode != EXIT_SUCCESS:
            LOG.critical(result.stderr.decode("utf-8").rstrip())
            raise CalledProcessError(result.returncode, final_command)

    def _build_config_input(self, case_id: str, fastq_path: Path, **flags) -> BalsamicConfigInput:
        case: Case = self.status_db.get_case_by_internal_id_strict(case_id)
        if self._all_samples_are_wgs(case):
            return self._build_wgs_config(case=case, fastq_path=fastq_path)
        else:
            return self._build_targeted_config(
                case=case, fastq_path=fastq_path, override_panel_bed=flags.get("panel_bed")
            )

    def _build_wgs_config(self, case: Case, fastq_path: Path) -> BalsamicConfigInput:
        patient_sex: SexOptions = self._get_patient_sex(case)
        return BalsamicConfigInputWGS(
            analysis_dir=self.root_dir,
            analysis_workflow=case.data_analysis,
            artefact_snv_observations=self.loqusdb_artefact_snv,
            artefact_sv_observations=self.artefact_sv_observations,
            balsamic_binary=self.balsamic_binary,
            balsamic_cache=self.cache_dir,
            cache_version=self.cache_version,
            cadd_annotations=self.cadd_path,
            cancer_germline_snv_observations=self.loqusdb_cancer_germline_snv,
            cancer_somatic_snv_observations=self.loqusdb_cancer_somatic_snv,
            cancer_somatic_sv_observations=self.loqusdb_cancer_somatic_sv,
            case_id=case.internal_id,
            clinical_snv_observations=self.loqusdb_clinical_snv,
            clinical_sv_observations=self.loqusdb_clinical_sv,
            conda_binary=self.conda_binary,
            conda_env=self.conda_env,
            fastq_path=fastq_path,
            gender=patient_sex,
            genome_interval=self.genome_interval_path,
            genome_version=GenomeVersion.HG19,
            gens_coverage_pon=self._get_gens_coverage_pon_file(patient_sex),
            gnomad_min_af5=self.gnomad_af5_path,
            normal_sample_name=self._get_normal_sample_id_from_paired_analysis(case),
            sentieon_install_dir=self.sentieon_licence_path,
            sentieon_license=self.sentieon_licence_server,
            swegen_snv=self.swegen_snv,
            swegen_sv=self.swegen_sv,
            tumor_sample_name=self._get_tumor_or_single_sample_id(case),
        )

    def _build_targeted_config(
        self, case: Case, fastq_path: Path, override_panel_bed: str | None
    ) -> BalsamicConfigInput:
        bed_version: BedVersion = self._get_bed_version(
            case=case, override_panel_bed=override_panel_bed
        )
        bed_file: Path = Path(self.bed_directory, bed_version.filename)
        patient_sex: SexOptions = self._get_patient_sex(case)
        return BalsamicConfigInputPanel(
            analysis_dir=self.root_dir,
            analysis_workflow=case.data_analysis,
            artefact_snv_observations=self.loqusdb_artefact_snv,
            balsamic_binary=self.balsamic_binary,
            balsamic_cache=self.cache_dir,
            cache_version=self.cache_version,
            cadd_annotations=self.cadd_path,
            cancer_germline_snv_observations=self.loqusdb_cancer_germline_snv,
            cancer_somatic_snv_observations=self.loqusdb_cancer_somatic_snv,
            cancer_somatic_snv_panel_observations=self.loqusdb_cancer_somatic_snv_panels.get(
                bed_version.bed_name
            ),
            cancer_somatic_sv_observations=self.loqusdb_cancer_somatic_sv,
            case_id=case.internal_id,
            clinical_snv_observations=self.loqusdb_clinical_snv,
            clinical_sv_observations=self.loqusdb_clinical_sv,
            conda_binary=self.conda_binary,
            conda_env=self.conda_env,
            fastq_path=fastq_path,
            gender=patient_sex,
            genome_version=GenomeVersion.HG19,
            gnomad_min_af5=self.gnomad_af5_path,
            normal_sample_name=self._get_normal_sample_id_from_paired_analysis(case),
            panel_bed=bed_file,
            pon_cnn=self._get_pon_file(bed_version.shortname),
            exome=self._all_samples_are_exome(case),
            sentieon_install_dir=self.sentieon_licence_path,
            sentieon_license=self.sentieon_licence_server,
            soft_filter_normal=self._is_case_paired_analysis(case),
            swegen_snv=self.swegen_snv,
            swegen_sv=self.swegen_sv,
            tumor_sample_name=self._get_tumor_or_single_sample_id(case),
        )

    @staticmethod
    def _all_samples_are_wgs(case: Case) -> bool:
        """Check if all samples in the case are WGS."""
        return all(
            sample.prep_category == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
            for sample in case.samples
        )

    @staticmethod
    def _all_samples_are_exome(case: Case) -> bool:
        """Check if all samples in the case are exome."""
        return all(
            sample.prep_category == SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
            for sample in case.samples
        )

    @staticmethod
    def _get_patient_sex(case) -> SexOptions:
        sample_sex: set[SexOptions] = {sample.sex for sample in case.samples}
        return sample_sex.pop()

    @staticmethod
    def _get_normal_sample_id_from_paired_analysis(case) -> str | None:
        """Return the internal id of the normal sample if the case is a paired analysis, otherwise None."""
        if len(case.samples) == 2:
            for sample in case.samples:
                if not sample.is_tumour:
                    return sample.internal_id

    @staticmethod
    def _get_tumor_or_single_sample_id(case) -> str:
        """
        Return the internal id of the tumour sample if the case is a paired analysis,
        otherwise return the internal id of the single sample.
        """
        if len(case.samples) == 1:
            return case.samples[0].internal_id
        for sample in case.samples:
            if sample.is_tumour:
                return sample.internal_id

    @staticmethod
    def _is_case_paired_analysis(case: Case) -> bool:
        return len(case.samples) == 2

    def _get_bed_version(self, case: Case, override_panel_bed: str | None) -> BedVersion:
        first_sample: Sample = case.samples[0]
        short_name: str = override_panel_bed or self.lims_api.get_capture_kit_strict(
            first_sample.internal_id
        )
        return self.status_db.get_bed_version_by_short_name_and_genome_version_strict(
            short_name=short_name, genome_version=BedVersionGenomeVersion.HG19
        )

    def _get_pon_file(self, bed_short_name: str | None) -> Path | None:
        if pon_file := self.panel_of_normals.get(bed_short_name):
            return pon_file
        else:
            LOG.info(f"No PON file found for bed file {pon_file}. Configuring without PON.")
            return None

    def _get_sample_config_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id, f"{case_id}.json")

    def _get_gens_coverage_pon_file(self, patient_sex: SexOptions) -> Path:
        """Return the corresponding PON file for WGS cases based on the patient's sex."""
        return (
            self.gens_coverage_male_path
            if patient_sex == SexOptions.MALE
            else self.gens_coverage_female_path
        )



from typing import Generic, TypeVar

from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.sample import Sample

SampleType = TypeVar("SampleType", bound=Sample)


class OrderWithSamples(Order, Generic[SampleType]):
    samples: list[SampleType]

    @property
    def enumerated_samples(self) -> enumerate[SampleType]:
        return enumerate(self.samples)



"""Raredisease Delivery Report API."""

from housekeeper.store.models import File

from cg.clients.chanjo2.models import CoverageMetrics
from cg.constants.housekeeper_tags import AnalysisTag
from cg.constants.report import (
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_RAREDISEASE_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_METADATA_RAREDISEASE_FIELDS,
    REQUIRED_SAMPLE_METADATA_RAREDISEASE_WGS_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_RAREDISEASE_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
)
from cg.constants.scout import ScoutUploadKey
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.meta.delivery_report.data_validators import get_million_read_pairs
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.analysis import AnalysisModel, NextflowAnalysis
from cg.models.delivery_report.metadata import RarediseaseSampleMetadataModel
from cg.models.delivery_report.report import CaseModel, ReportRequiredFields, ScoutVariantsFiles
from cg.models.delivery_report.sample import SampleModel
from cg.models.raredisease.raredisease import RarediseaseQCMetrics
from cg.store.models import Case, Sample


class RarediseaseDeliveryReportAPI(DeliveryReportAPI):
    """API to create Raredisease delivery reports."""

    def __init__(self, analysis_api: RarediseaseAnalysisAPI):
        super().__init__(analysis_api=analysis_api)

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: NextflowAnalysis
    ) -> RarediseaseSampleMetadataModel:
        """Return Raredisease sample metadata to include in the report."""
        sample_metrics: RarediseaseQCMetrics = analysis_metadata.sample_metrics[sample.internal_id]
        gene_ids: list[int] = self.analysis_api.get_gene_ids_from_scout(case.panels)
        coverage_metrics: CoverageMetrics | None = self.analysis_api.get_sample_coverage(
            case_id=case.internal_id, sample_id=sample.internal_id, gene_ids=gene_ids
        )
        return RarediseaseSampleMetadataModel(
            bait_set=self.lims_api.capture_kit(sample.internal_id),
            duplicates=round(sample_metrics.percent_duplication * 100, 2),
            initial_qc=self.lims_api.has_sample_passed_initial_qc(sample.internal_id),
            mapped_reads=round((sample_metrics.picard_pct_pf_reads_aligned) * 100, 2),
            mean_target_coverage=coverage_metrics.mean_coverage if coverage_metrics else None,
            million_read_pairs=get_million_read_pairs(sample.reads),
            pct_10x=coverage_metrics.coverage_completeness_percent if coverage_metrics else None,
            sex=sample_metrics.predicted_sex_sex_check,
        )

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel = None
    ) -> bool:
        """
        Return whether the Raredisease delivery report is accredited.
        This method evaluates the accreditation status of each sample's application.
        """
        return all(sample.application.accredited for sample in samples)

    def get_scout_variants_files(self, case_id: str) -> ScoutVariantsFiles:
        """Return Raredisease files that will be uploaded to Scout."""
        snv_vcf: str | None = self.get_scout_uploaded_file_from_hk(
            case_id=case_id, scout_key=ScoutUploadKey.VCF_SNV
        )
        sv_vcf: str | None = self.get_scout_uploaded_file_from_hk(
            case_id=case_id, scout_key=ScoutUploadKey.VCF_SV
        )
        vcf_str_file: File | None = self.housekeeper_api.get_latest_file(
            bundle=case_id, tags=[AnalysisTag.VCF_STR, case_id]
        )
        vcf_str: str | None = vcf_str_file.full_path if vcf_str_file else None
        smn_tsv: str | None = self.get_scout_uploaded_file_from_hk(
            case_id=case_id, scout_key=ScoutUploadKey.SMN_TSV
        )
        return ScoutVariantsFiles(
            snv_vcf=snv_vcf,
            sv_vcf=sv_vcf,
            vcf_str=vcf_str,
            smn_tsv=smn_tsv,
        )

    @staticmethod
    def get_sample_metadata_required_fields(case: CaseModel) -> dict:
        """Return sample metadata required fields associated to a specific sample."""
        required_sample_metadata_fields = {}
        for sample in case.samples:
            required_fields = (
                REQUIRED_SAMPLE_METADATA_RAREDISEASE_WGS_FIELDS
                if SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
                in sample.application.prep_category.lower()
                else REQUIRED_SAMPLE_METADATA_RAREDISEASE_FIELDS
            )
            required_sample_metadata_fields.update({sample.id: required_fields})
        return required_sample_metadata_fields

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields for Raredisease."""
        report_required_fields = ReportRequiredFields(
            applications=self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            case=REQUIRED_CASE_FIELDS,
            customer=REQUIRED_CUSTOMER_FIELDS,
            data_analysis=REQUIRED_DATA_ANALYSIS_RAREDISEASE_FIELDS,
            metadata=self.get_sample_metadata_required_fields(case=case),
            methods=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            report=REQUIRED_REPORT_FIELDS,
            samples=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_RAREDISEASE_FIELDS
            ),
            timestamps=self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
        )
        return report_required_fields.model_dump()



"""Module to read PNG image files."""

import base64
from pathlib import Path

from cg.constants import FileExtensions
from cg.io.validate_path import validate_file_suffix


def read_png(file_path: Path) -> str:
    """Return base64 encoding of a PNG image."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtensions.PNG)
    with open(file_path, "rb") as png_file:
        encoded_string: str = base64.b64encode(png_file.read()).decode("utf-8")
    return f"data:image/png;base64,{encoded_string}"



from typing import Any

from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.case_sample_errors import CaseSampleError
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.errors.sample_errors import SampleError
from cg.services.orders.validation.errors.validation_errors import ValidationErrors


def create_order_validation_response(raw_order: dict, errors: ValidationErrors) -> dict:
    """Ensures each field in the order looks like: {value: raw value, errors: [errors]}"""
    wrap_fields(raw_order)
    map_errors_to_order(order=raw_order, errors=errors)
    return raw_order


def map_errors_to_order(order: dict, errors: ValidationErrors) -> None:
    map_order_errors(order=order, errors=errors.order_errors)
    map_case_errors(order=order, errors=errors.case_errors)
    map_case_sample_errors(order=order, errors=errors.case_sample_errors)
    map_sample_errors(order=order, errors=errors.sample_errors)


def map_order_errors(order: dict, errors: list[OrderError]) -> None:
    for error in errors:
        add_error(entity=order, field=error.field, message=error.message)


def map_case_errors(order: dict, errors: list[CaseError]) -> None:
    for error in errors:
        case: dict = get_case(order=order, index=error.case_index)
        add_error(entity=case, field=error.field, message=error.message)


def map_case_sample_errors(order: dict, errors: list[CaseSampleError]) -> None:
    for error in errors:
        case: dict = get_case(order=order, index=error.case_index)
        sample: dict = get_case_sample(case=case, index=error.sample_index)
        add_error(entity=sample, field=error.field, message=error.message)


def map_sample_errors(order: dict, errors: list[SampleError]) -> None:
    for error in errors:
        sample: dict = get_sample(order=order, index=error.sample_index)
        add_error(entity=sample, field=error.field, message=error.message)


def add_error(entity: dict, field: str, message: str) -> None:
    if not entity.get(field):
        set_field(entity=entity, field=field, value=None)
    if field in {"sample_errors", "warnings"}:
        # Special handling since the 'value' corresponds to whether it is set
        entity[field]["value"] = True
    entity[field]["errors"].append(message)


def get_case(order: dict, index: int) -> dict:
    return order["cases"][index]


def get_case_sample(case: dict, index: int) -> dict:
    return case["samples"][index]


def get_sample(order: dict, index: int) -> dict:
    return order["samples"][index]


def wrap_fields(raw_order: dict) -> None:
    wrap_order_fields(raw_order)
    if raw_order.get("cases"):
        wrap_case_and_sample_fields(raw_order)
    else:
        wrap_sample_fields(raw_order["samples"])


def wrap_order_fields(raw_order: dict) -> None:
    for field, value in raw_order.items():
        if field not in {"cases", "samples"}:
            set_field(entity=raw_order, field=field, value=value)


def wrap_case_and_sample_fields(raw_order: dict) -> None:
    for case in raw_order["cases"]:
        wrap_case_fields(case)
        wrap_sample_fields(case["samples"])


def wrap_case_fields(case: dict) -> None:
    for field, value in case.items():
        if field != "samples":
            set_field(entity=case, field=field, value=value)
    set_field(entity=case, field="sample_errors", value=False)


def wrap_sample_fields(samples: list[dict]) -> None:
    for sample in samples:
        for field, value in sample.items():
            set_field(entity=sample, field=field, value=value)
        set_field(entity=sample, field="warnings", value=False)


def set_field(entity: dict, field: str, value: Any) -> None:
    entity[field] = {"value": value, "errors": []}



"""Module for teh Pacbio sequenicng metrics parsing service."""

import logging
from pathlib import Path

from pydantic import ValidationError

from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.run_devices.abstract_classes import PostProcessingMetricsParser
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingParsingError,
    PostProcessingRunFileManagerError,
)
from cg.services.run_devices.pacbio.metrics_parser.models import (
    BarcodeMetrics,
    ControlMetrics,
    MetadataMetrics,
    PacBioMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SampleMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.services.run_devices.pacbio.metrics_parser.utils import (
    get_parsed_metadata_file,
    get_parsed_metrics_from_file_name,
    get_parsed_sample_metrics,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager

LOG = logging.getLogger(__name__)


class PacBioMetricsParser(PostProcessingMetricsParser):
    """Class for parsing PacBio sequencing metrics."""

    def __init__(self, file_manager: PacBioRunFileManager):
        self.file_manager: PacBioRunFileManager = file_manager

    @handle_post_processing_errors(
        to_except=(FileNotFoundError, ValidationError, PostProcessingRunFileManagerError),
        to_raise=PostProcessingParsingError,
    )
    def parse_metrics(self, run_data: PacBioRunData) -> PacBioMetrics:
        """Return all the relevant PacBio metrics parsed in a single Pydantic object."""
        metrics_files: list[Path] = self.file_manager.get_files_to_parse(run_data)
        read_metrics: ReadMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.CCS_REPORT_SUFFIX
        )
        control_metrics: ControlMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.CONTROL_REPORT
        )
        productivity_metrics: ProductivityMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.LOADING_REPORT
        )
        polymerase_metrics: PolymeraseMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.RAW_DATA_REPORT
        )
        dataset_metrics: SmrtlinkDatasetsMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT
        )
        barcodes_metrics: BarcodeMetrics = get_parsed_metrics_from_file_name(
            metrics_files=metrics_files, file_name=PacBioDirsAndFiles.BARCODES_REPORT
        )
        metadata_metrics: MetadataMetrics = get_parsed_metadata_file(metrics_files)
        sample_metrics: list[SampleMetrics] = get_parsed_sample_metrics(metrics_files)
        LOG.debug(f"All metrics parsed for run {run_data.run_id}")
        return PacBioMetrics(
            read=read_metrics,
            control=control_metrics,
            productivity=productivity_metrics,
            polymerase=polymerase_metrics,
            dataset_metrics=dataset_metrics,
            barcodes=barcodes_metrics,
            metadata=metadata_metrics,
            samples=sample_metrics,
        )



import logging
from pathlib import Path

from cg.meta.workflow.microsalt.quality_controller.models import (
    CaseQualityResult,
    SampleQualityResult,
)

LOG = logging.getLogger(__name__)


class ResultLogger:
    @staticmethod
    def log_results(
        samples: list[SampleQualityResult], case: CaseQualityResult, report: Path
    ) -> None:
        if case.passes_qc:
            LOG.info(f"QC passed, see {report} for details.")
        else:
            message = get_case_fail_message(case)
            LOG.warning(message)

        message = sample_result_message(samples)
        LOG.info(message)

    @staticmethod
    def log_sample_result(result: SampleQualityResult) -> None:
        control_message = "Control sample " if result.is_control else ""
        if result.passes_qc:
            message = f"{control_message}{result.sample_id} passed QC."
            LOG.info(message)
        else:
            message = f"{control_message}{result.sample_id} failed QC."
            LOG.warning(message)

    @staticmethod
    def log_case_result(result: CaseQualityResult) -> None:
        if not result.passes_qc:
            LOG.warning("Case failed QC.")


def get_case_fail_message(case: CaseQualityResult) -> str:
    fail_reasons = []

    if not case.control_passes_qc:
        fail_reasons.append("The negative control sample failed QC.\n")
    fail_message = "QC failed.\n"
    return fail_message + "".join(fail_reasons)


def sample_result_message(samples: list[SampleQualityResult]) -> str:
    failed_samples: list[SampleQualityResult] = get_failed_results(samples)
    passed_samples: list[SampleQualityResult] = get_passed_results(samples)

    failed_count: int = len(failed_samples)
    passed_count: int = len(passed_samples)
    total_count: int = len(samples)

    return f"Sample results: {failed_count} failed, {passed_count} passed, {total_count} total."


def get_failed_results(samples: list[SampleQualityResult]) -> list[SampleQualityResult]:
    return [result for result in samples if not result.passes_qc]


def get_passed_results(samples: list[SampleQualityResult]) -> list[SampleQualityResult]:
    return [result for result in samples if result.passes_qc]



from cg.models.qc_metrics import QCMetrics


class TomteQCMetrics(QCMetrics):
    """Tomte QC metrics."""

    after_filtering_gc_content: float
    after_filtering_q20_rate: float
    after_filtering_q30_rate: float
    after_filtering_read1_mean_length: float
    before_filtering_total_reads: float
    median_5prime_to_3prime_bias: float
    pct_adapter: float
    pct_duplication: float
    pct_intergenic_bases: float
    pct_intronic_bases: float
    pct_mrna_bases: float
    pct_ribosomal_bases: float
    pct_surviving: float
    uniquely_mapped_percent: float



from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.order_types.rna_fusion.models.sample import RNAFusionSample


class RNAFusionCase(Case[RNAFusionSample]):
    cohorts: list[str] | None = None
    synopsis: str | None = None



from cg.services.sample_run_metrics_service.dtos import IlluminaSequencingMetricsDTO
from cg.store.models import IlluminaSampleSequencingMetrics


def create_metrics_dto(
    metrics: list[IlluminaSampleSequencingMetrics] | None,
) -> list[IlluminaSequencingMetricsDTO]:

    if not metrics:
        return []

    parsed_metrics = []

    for metric in metrics:
        parsed_metric = IlluminaSequencingMetricsDTO(
            flow_cell_name=metric.instrument_run.device.internal_id,
            flow_cell_lane_number=metric.flow_cell_lane,
            sample_internal_id=metric.sample.internal_id,
            sample_total_reads_in_lane=metric.total_reads_in_lane,
            sample_base_percentage_passing_q30=metric.base_passing_q30_percent,
        )
        parsed_metrics.append(parsed_metric)
    return parsed_metrics



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any


class ParamsFileCreator(ABC):

    def __init__(self, params: str):
        self.params = Path(params)

    @abstractmethod
    def create(self, case_id: str, file_path: Path, sample_sheet_path: Path) -> Any:
        pass



import logging
import os
import shutil
from pathlib import Path

from cg.apps.demultiplex.sample_sheet.utils import add_and_include_sample_sheet_path_to_housekeeper
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.constants import Workflow
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.priority import TrailblazerPriority
from cg.constants.tb import AnalysisStatus, AnalysisType
from cg.io.csv import read_csv, write_csv

LOG = logging.getLogger(__name__)

NANOPORE_SEQUENCING_SUMMARY_PATTERN: str = r"final_summary_*.txt"


def is_flow_cell_demultiplexed(analysis_directory: Path) -> bool:
    """Determine whether the flow cell has been demultiplexed for the latest analysis for a Novaseqx flow cell."""
    return Path(
        analysis_directory,
        DemultiplexingDirsAndFiles.DATA,
        DemultiplexingDirsAndFiles.ANALYSIS_COMPLETED,
    ).exists()


def is_flow_cell_in_demultiplexed_runs(flow_cell_name: str, demultiplexed_runs: Path) -> bool:
    """Determine whether the flow cell is in the demultiplexed runs directory."""
    return Path(demultiplexed_runs, flow_cell_name).exists()


def get_latest_analysis_path(flow_cell_dir: Path) -> Path | None:
    """
    Get the latest analysis directory for a Novaseqx flow cell.
    The latest analysis directory is the one with the highest integer name.
    """
    analysis_path: Path = Path(flow_cell_dir, DemultiplexingDirsAndFiles.ONBOARD_ANALYSIS)
    if not analysis_path.exists():
        return None
    analysis_versions: list[Path] = get_sorted_analysis_versions(analysis_path)
    return analysis_versions[0] if analysis_versions else None


def get_sorted_analysis_versions(analysis_path: Path) -> list[Path]:
    """Get a sorted list of analysis version paths for a Novaseqx flow cell."""

    def sort_by_name(version: Path) -> int:
        return int(version.name)

    analysis_versions_paths: list[Path] = [
        version_path for version_path in analysis_path.iterdir() if version_path.is_dir()
    ]
    return sorted(analysis_versions_paths, key=sort_by_name, reverse=True)


def hardlink_flow_cell_analysis_data(flow_cell_dir: Path, demultiplexed_runs_dir: Path) -> None:
    """Create hardlinks to the latest version of the analysis data for a Novaseqx flow cell."""
    analysis_path: Path = get_latest_analysis_path(flow_cell_dir)
    analysis_data_path: Path = Path(analysis_path, DemultiplexingDirsAndFiles.DATA)
    demultiplexed_runs_dir_flow_cell_dir = Path(demultiplexed_runs_dir, flow_cell_dir.name)
    hardlink_tree(src=analysis_data_path, dst=demultiplexed_runs_dir_flow_cell_dir)


def mark_as_demultiplexed(flow_cell_dir: Path) -> None:
    """Create the demux_complete.txt file in the specified flow cell directory."""
    Path(flow_cell_dir, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()


def hardlink_tree(src: Path, dst: Path) -> None:
    """Create hardlinks to the specified source directory in the specified destination directory."""
    shutil.copytree(src=src, dst=dst, copy_function=os.link)


def is_ready_to_copy_to_demultiplexed_runs(
    flow_cell_dir: Path, demultiplexed_runs_dir: Path
) -> bool:
    """
    Determine whether a NovaSeqX flow cell is ready to be hard-linked into the demultiplexed-runs directory.
    All three conditions must be met:
    - Sync of data is confirmed (CopyComplete.txt at the flow cell root, written by confirm-sequencing-run-sync)
    - on-instrument BCLConvert demultiplexing has completed
    - the flow cell has not already been copied to demultiplexed-runs
    """
    analysis_path: Path = get_latest_analysis_path(flow_cell_dir)

    if not analysis_path:
        LOG.debug(f"No analysis path found for flow cell {flow_cell_dir.name}.")
        return False

    flow_cell_is_ready: bool = True

    if not is_flow_cell_sync_confirmed(flow_cell_dir):
        LOG.debug(f"Sync has not been confirmed for flow cell {flow_cell_dir.name}.")
        flow_cell_is_ready = False
    if not is_flow_cell_demultiplexed(analysis_path):
        LOG.debug(f"Flow cell {flow_cell_dir.name} has not been demultiplexed.")
        flow_cell_is_ready = False

    if is_flow_cell_in_demultiplexed_runs(
        flow_cell_name=flow_cell_dir.name, demultiplexed_runs=demultiplexed_runs_dir
    ):
        LOG.debug(f"Flow cell {flow_cell_dir.name} is already in the demultiplexed runs directory.")
        flow_cell_is_ready = False

    return flow_cell_is_ready


def get_existing_manifest_file(source_directory: Path) -> Path | None:
    """Returns the first existing manifest file in the source directory."""
    manifest_file_paths = [
        Path(source_directory, DemultiplexingDirsAndFiles.ILLUMINA_FILE_MANIFEST),
        Path(source_directory, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST),
    ]
    for file_path in manifest_file_paths:
        if file_path.exists():
            return file_path


def are_all_files_synced(files_at_source: list[Path], target_directory: Path) -> bool:
    """Checks if all relevant files in the source are present in the target directory."""
    for file in files_at_source:
        target_file_path = Path(target_directory, file)
        if is_file_relevant_for_demultiplexing(file) and not target_file_path.exists():
            LOG.info(f"File: {file}, has not been transferred from source to {target_directory}")
            return False
    return True


def is_syncing_complete(source_directory: Path, target_directory: Path) -> bool:
    """Returns whether all relevant files for demultiplexing have been synced from the source to the target."""
    existing_manifest_file: Path | None = get_existing_manifest_file(source_directory)

    if not existing_manifest_file:
        LOG.debug(f"{source_directory} does not contain a manifest file. Skipping.")
        return False

    files_at_source: list[Path] = parse_manifest_file(existing_manifest_file)

    analysis_path: Path | None = get_latest_analysis_path(source_directory)
    if analysis_path:
        analysis_manifest: Path | None = get_existing_manifest_file(analysis_path)
        if analysis_manifest:
            analysis_prefix: Path = analysis_path.relative_to(source_directory)
            analysis_files: list[Path] = parse_manifest_file(analysis_manifest)
            files_at_source.extend(Path(analysis_prefix, f) for f in analysis_files)

    return are_all_files_synced(files_at_source=files_at_source, target_directory=target_directory)


def is_manifest_file_required(flow_cell_dir: Path) -> bool:
    """Returns whether a flow cell directory needs a manifest file."""
    illumina_manifest_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.ILLUMINA_FILE_MANIFEST)
    custom_manifest_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)
    copy_complete_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.COPY_COMPLETE)
    sequencing_finished: bool = copy_complete_file.exists() or is_nanopore_sequencing_complete(
        flow_cell_dir
    )
    return (
        not any((illumina_manifest_file.exists(), custom_manifest_file.exists()))
        and sequencing_finished
    )


def create_manifest_file(flow_cell_dir_name: Path) -> Path:
    """Creates a tab separated file containing the paths of all files in the given
    directory and any subdirectories."""
    files_in_directory: list[list[str]] = []
    for subdir, _, files in os.walk(flow_cell_dir_name):
        subdir = Path(subdir).relative_to(flow_cell_dir_name)
        files_in_directory.extend([Path(subdir, file).as_posix()] for file in files)
    LOG.info(
        f"Writing manifest file to {Path(flow_cell_dir_name, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)}"
    )
    output_path = Path(flow_cell_dir_name, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)
    write_csv(
        content=files_in_directory,
        file_path=output_path,
        delimiter="\t",
    )
    return output_path


def is_flow_cell_sync_confirmed(target_flow_cell_dir: Path) -> bool:
    return Path(target_flow_cell_dir, DemultiplexingDirsAndFiles.COPY_COMPLETE).exists()


def get_nanopore_summary_file(flow_cell_directory: Path) -> bool | None:
    """Returns the summary file for a Nanopore run if found."""
    try:
        file = Path(next(flow_cell_directory.glob(NANOPORE_SEQUENCING_SUMMARY_PATTERN)))
    except StopIteration:
        file = None
    return file


def is_nanopore_sequencing_complete(flow_cell_directory: Path) -> bool:
    """Returns whether the sequencing is complete for a Nanopore run."""
    return bool(get_nanopore_summary_file(flow_cell_directory))


def parse_manifest_file(manifest_file: Path) -> list[Path]:
    """Returns a list with the first entry of each row of the given TSV file."""
    LOG.debug(f"Parsing manifest file: {manifest_file}")
    files: list[list[str]] = read_csv(file_path=manifest_file, delimiter="\t")
    return [Path(file[0]) for file in files]


def is_file_relevant_for_demultiplexing(file: Path) -> bool:
    """Returns whether a file is relevant for demultiplexing."""
    relevant_directories = [DemultiplexingDirsAndFiles.INTER_OP, DemultiplexingDirsAndFiles.DATA]
    for relevant_directory in relevant_directories:
        if relevant_directory in file.parts:
            return True
    return False


def link_onboard_demultiplexed_flow_cell(
    sequencing_run_dir: Path,
    demultiplexed_runs_dir: Path,
    flow_cell_id: str,
    hk_api: HousekeeperAPI,
    tb_api: TrailblazerAPI,
) -> None:
    """Hard-link an on-instrument demultiplexed NovaSeqX flow cell into demultiplexed-runs and notify downstream systems."""
    hardlink_flow_cell_analysis_data(
        flow_cell_dir=sequencing_run_dir, demultiplexed_runs_dir=demultiplexed_runs_dir
    )
    demultiplexed_runs_flow_cell_dir: Path = Path(demultiplexed_runs_dir, sequencing_run_dir.name)
    mark_as_demultiplexed(demultiplexed_runs_flow_cell_dir)

    analysis_path: Path = get_latest_analysis_path(sequencing_run_dir)
    add_and_include_sample_sheet_path_to_housekeeper(
        flow_cell_directory=Path(analysis_path, DemultiplexingDirsAndFiles.DATA),
        flow_cell_name=flow_cell_id,
        hk_api=hk_api,
    )
    tb_api.add_pending_analysis(
        case_id=flow_cell_id,
        analysis_type=AnalysisType.OTHER,
        config_path="",
        out_dir=demultiplexed_runs_flow_cell_dir.as_posix(),
        priority=TrailblazerPriority.HIGH,
        workflow=Workflow.DEMULTIPLEX,
    )
    tb_api.set_analysis_status(case_id=flow_cell_id, status=AnalysisStatus.COMPLETED)



import logging
from http import HTTPStatus

import flask
import requests
from flask import Blueprint, jsonify, request

from cg.exc import AnalysisDoesNotExistError, TrailblazerAPIHTTPError
from cg.server.endpoints.utils import before_request
from cg.server.ext import db, mark_as_delivered_service
from cg.store.models import Analysis

DELIVER_BLUEPRINT = Blueprint("deliver", __name__, url_prefix="/api/v1")
DELIVER_BLUEPRINT.before_request(before_request)

LOG = logging.getLogger(__name__)


@DELIVER_BLUEPRINT.route("/deliver", methods=["POST"])
def deliver_analyses():
    """
    Marks the analyses as delivered.
    - Adds a date stamp to the 'delivered_at' field for the relevant samples.
    - Calls Trailblazer to mark the Trailblazer analyses entries as delivered.
    """
    try:
        trailblazer_ids: list[int] = request.json["trailblazer_ids"]  # type: ignore None is handled
        analyses = []
        for trailblazer_id in trailblazer_ids:
            analysis: Analysis = db.get_analysis_by_trailblazer_id(trailblazer_id)
            analyses.append(analysis)
        analyses_response: requests.Response = mark_as_delivered_service.mark_analyses(analyses)
    except AnalysisDoesNotExistError as error:
        LOG.error(str(error))
        return jsonify(message=str(error)), HTTPStatus.BAD_REQUEST
    except (KeyError, TypeError):
        return flask.Response(status=HTTPStatus.BAD_REQUEST)
    except TrailblazerAPIHTTPError as error:
        LOG.error(f"Error in Trailblazer: {str(error)} - rolling back changes in StatusDB.")
        db.rollback()
        return jsonify(message="Error when calling Trailblazer"), HTTPStatus.BAD_GATEWAY
    finally:
        db.commit_to_store()
    return flask.Response(response=analyses_response.content, status=HTTPStatus.OK)



from abc import abstractmethod, ABC

from cg.services.deliver_files.tag_fetcher.models import DeliveryFileTags


class FetchDeliveryFileTagsService(ABC):
    """
    Abstract class that encapsulates the logic required for fetching tags for files to deliver.
    """

    @abstractmethod
    def fetch_tags(self, workflow: str) -> DeliveryFileTags:
        """Fetch the tags for the files to deliver."""
        pass



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.rna_fusion.constants import RNAFusionDeliveryType
from cg.services.orders.validation.order_types.rna_fusion.models.case import RNAFusionCase
from cg.services.orders.validation.order_types.rna_fusion.models.sample import RNAFusionSample


class RNAFusionOrder(OrderWithCases[RNAFusionCase, RNAFusionSample]):
    delivery_type: RNAFusionDeliveryType



import logging

from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants.priority import SlurmQos
from cg.constants.slurm import (
    SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES,
    SLURM_UPLOAD_MAX_HOURS,
    SLURM_UPLOAD_MEMORY,
    SLURM_UPLOAD_TASKS,
)
from cg.models.slurm.sbatch import Sbatch
from cg.services.slurm_service.slurm_service import SlurmService
from cg.services.slurm_upload_service.slurm_upload_config import SlurmUploadConfig
from cg.services.slurm_upload_service.utils import get_quality_of_service

LOG = logging.getLogger(__name__)


class SlurmUploadService:
    def __init__(
        self,
        config: SlurmUploadConfig,
        slurm_service: SlurmService,
        trailblazer_api: TrailblazerAPI,
    ):
        self.slurm_service = slurm_service
        self.trailblazer_api = trailblazer_api
        self.config = config

    def upload(self, upload_command: str, job_name: str, case_id: str):
        LOG.debug(f"Uploading case {case_id} to via SLURM with command: {upload_command}")
        analysis: TrailblazerAnalysis = self.trailblazer_api.get_latest_completed_analysis(case_id)
        job_name = f"{job_name}_{analysis.id}"
        job_config: Sbatch = self._get_job_config(command=upload_command, job_name=job_name)
        slurm_id: int = self.slurm_service.submit_job(job_config)
        self.trailblazer_api.add_upload_job_to_analysis(slurm_id=slurm_id, analysis_id=analysis.id)

    def _get_job_config(self, command: str, job_name: str) -> Sbatch:
        quality_of_service: SlurmQos = get_quality_of_service(self.config.account)
        return Sbatch(
            account=self.config.account,
            commands=command,
            email=self.config.email,
            exclude=SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES,
            hours=SLURM_UPLOAD_MAX_HOURS,
            job_name=job_name,
            log_dir=self.config.log_dir,
            memory=SLURM_UPLOAD_MEMORY,
            number_tasks=SLURM_UPLOAD_TASKS,
            quality_of_service=quality_of_service,
        )



NARROW_NO_BREAK_SPACE = "\u202f"

SI_PREFIXES = [(1e12, "T", 2), (1e9, "G", 2), (1e6, "M", 2), (1e3, "k", 2), (1, "", 0)]


class Si:
    """
    Formats numbers according to SI standards.

    References:
        BIPM (2019). The International System of Units (SI Brochure), 9th edition.
        Section 5.4.4: Formatting numbers. https://www.bipm.org/en/publications/si-brochure/
    """

    @staticmethod
    def prefix(value: int | float, unit: str) -> str:
        """
        Formats numbers according to selected SI prefixes.
        """
        value = Si._integer_convert(value)
        for threshold, prefix, decimals in SI_PREFIXES:
            if value >= threshold:
                scaled_value = value / threshold
                return f"{scaled_value:.{decimals}f}{NARROW_NO_BREAK_SPACE}{prefix}{unit}"
        return f"{value}{NARROW_NO_BREAK_SPACE}{unit}"

    @staticmethod
    def group_digits(value: int | float) -> str:
        """
        Format according to SI group digits, section 5.4.4 of SI Brochure, 9th edition
        """
        value = Si._integer_convert(value)
        return f"{value:,d}".replace(",", NARROW_NO_BREAK_SPACE)

    @staticmethod
    def _integer_convert(value: int | float) -> int:
        """
        Convert a number to an integer using SI-compliant round-half-to-even.
        """
        return int(round(value))



import logging
from pathlib import Path
import re
from cg.apps.lims import LimsAPI
from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.files.abstract import FileFormatter
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.files.concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.files.sample_service import FileManager

LOG = logging.getLogger(__name__)


class MutantFileFormatter(FileFormatter):
    """
    Formatter for file to deliver or upload for the Mutant workflow.
    Args:
        lims_api: The LIMS API
        file_formatter: The SampleFileConcatenationFormatter. This is used to format the files and concatenate the fastq files.
        file_manager: The FileManager

    """

    def __init__(
        self,
        lims_api: LimsAPI,
        file_formatter: SampleFileConcatenationFormatter,
        file_manager: FileManager,
    ):
        self.lims_api: LimsAPI = lims_api
        self.file_formatter: SampleFileConcatenationFormatter = file_formatter
        self.file_manager = file_manager

    def format_files(
        self, moved_files: list[SampleFile], delivery_path: Path
    ) -> list[FormattedFile]:
        """
        Format the mutant files to deliver and return the formatted files.
        args:
            moved_files: The sample files to format
            delivery_path: The path to deliver the files

        """
        LOG.debug("[FORMAT SERVICE] Formatting and concatenating mutant files")
        formatted_files: list[FormattedFile] = self.file_formatter.format_files(
            moved_files=moved_files, delivery_path=delivery_path
        )
        appended_formatted_files: list[FormattedFile] = self._add_lims_metadata_to_file_name(
            formatted_files=formatted_files, sample_files=moved_files
        )
        unique_formatted_files: list[FormattedFile] = self._filter_unique_path_combinations(
            appended_formatted_files
        )
        for unique_files in unique_formatted_files:
            self.file_manager.rename_file(
                src=unique_files.original_path, dst=unique_files.formatted_path
            )
        return unique_formatted_files

    @staticmethod
    def _is_concatenated_file(file_path: Path) -> bool:
        """Check if the file is a concatenated file.
        Returns True if the file is a concatenated file, otherwise False.
        regex pattern: *._[1,2].fastq.gz
        *. is the sample id
        _[1,2] is the read direction
        .fastq.gz is the file extension
        args:
            file_path: The file path to check
        """
        pattern = ".*_[1,2].fastq.gz"
        return re.fullmatch(pattern, file_path.name) is not None

    def _add_lims_metadata_to_file_name(
        self, formatted_files: list[FormattedFile], sample_files: list[SampleFile]
    ) -> list[FormattedFile]:
        """
        This functions adds the region and lab code to the file name of the formatted files.
        Note: The region and lab code is fetched from LIMS using the sample id. It is required for delivery of the files.
        This should only be done for concatenated fastq files.

        args:
            formatted_files: The formatted files to add the metadata to
            sample_files: The sample files to get the metadata from
        """
        appended_formatted_files: list[FormattedFile] = []
        for formatted_file in formatted_files:
            if self._is_concatenated_file(formatted_file.formatted_path):
                sample_id: str = self._get_sample_id_by_original_path(
                    original_path=formatted_file.original_path, sample_files=sample_files
                )
                lims_meta_data = self.lims_api.get_sample_region_and_lab_code(sample_id)

                new_original_path: Path = formatted_file.formatted_path
                new_formatted_path = Path(
                    formatted_file.formatted_path.parent,
                    f"{lims_meta_data}{formatted_file.formatted_path.name}",
                )
                appended_formatted_files.append(
                    FormattedFile(
                        original_path=new_original_path, formatted_path=new_formatted_path
                    )
                )
            else:
                appended_formatted_files.append(formatted_file)
        return appended_formatted_files

    @staticmethod
    def _get_sample_id_by_original_path(original_path: Path, sample_files: list[SampleFile]) -> str:
        """Get the sample id by the original path of the sample file.
        args:
            original_path: The original path of the sample file
            sample_files: The list of sample files to search in
        """
        for sample_file in sample_files:
            if sample_file.file_path == original_path:
                return sample_file.sample_id
        raise ValueError(f"Could not find sample file with path {original_path}")

    @staticmethod
    def _filter_unique_path_combinations(
        formatted_files: list[FormattedFile],
    ) -> list[FormattedFile]:
        """
        Filter out duplicates from the formatted files list.

        note:
            During fastq concatenation Sample_L1_R1 and Sample_L2_R1 files are concatenated
            and moved to the same file Concat_Sample. This mean that there can be multiple entries
            for the same concatenated file in the formatted_files list
            coming from the SampleFileConcatenationService.
            This function filters out the duplicates to avoid moving the same file multiple times
            which would result in an error the second time since the files is no longer in the original path.

        args:
            formatted_files: The formatted files to filter
        """
        unique_combinations = set()
        unique_files: list[FormattedFile] = []
        for formatted_file in formatted_files:
            combination = (formatted_file.original_path, formatted_file.formatted_path)
            if combination not in unique_combinations:
                unique_combinations.add(combination)
                unique_files.append(formatted_file)
        return unique_files



"""Module for demultiplexing CLI."""

import logging

import rich_click as click

from cg.cli.demultiplex.demux import (
    confirm_sequencing_run_sync,
    create_manifest_files,
    demultiplex_all,
    demultiplex_sequencing_run,
    link_onboard_demultiplexed_flow_cells,
)
from cg.cli.demultiplex.finish import finish_group
from cg.cli.demultiplex.sample_sheet import sample_sheet_commands
from cg.cli.utils import CLICK_CONTEXT_SETTINGS

LOG = logging.getLogger(__name__)


@click.group(name="demultiplex", context_settings=CLICK_CONTEXT_SETTINGS)
def demultiplex_cmd_group():
    """Command group for the demultiplex CLI."""
    LOG.info("Running cg demultiplex.")


demultiplex_cmd_group: click.Group
for sub_cmd in [
    create_manifest_files,
    confirm_sequencing_run_sync,
    demultiplex_sequencing_run,
    demultiplex_all,
    finish_group,
    link_onboard_demultiplexed_flow_cells,
    sample_sheet_commands,
]:
    demultiplex_cmd_group.add_command(sub_cmd)



from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.order_types.mip_rna.models.sample import MIPRNASample


class MIPRNACase(Case[MIPRNASample]):
    cohorts: list[str] | None = None
    synopsis: str | None = None



import logging
from datetime import datetime

from requests import Response

from cg.apps.tb.api import TrailblazerAPI
from cg.constants import Workflow
from cg.store.models import Analysis, Case, CaseSample, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class MarkAsDeliveredService:
    def __init__(self, status_db: Store, trailblazer_api: TrailblazerAPI) -> None:
        self.status_db = status_db
        self.trailblazer_api = trailblazer_api

    def mark_analyses(self, analyses: list[Analysis]) -> Response:
        """Mark samples as delivered in StatusDB and the analysis as delivered in Trailblazer."""
        trailblazer_ids = []
        for analysis in analyses:
            self._mark_samples_in_analysis(analysis)
            trailblazer_ids.append(analysis.trailblazer_id)
        return self.trailblazer_api.mark_analyses_as_delivered(trailblazer_ids)

    def _mark_samples_in_analysis(self, analysis: Analysis) -> None:
        case: Case = analysis.case
        LOG.info(f"Marking samples as delivered for case {case.internal_id}")
        for case_sample in case.links:
            if self._should_sample_be_delivered(case_sample):
                case_sample.sample.delivered_at = datetime.now()

    def _should_sample_be_delivered(self, case_sample: CaseSample) -> bool:
        return (
            case_sample.should_deliver_sample
            and not case_sample.sample.delivered_at
            and self._passes_on_reads(case_sample)
        )

    @staticmethod
    def _passes_on_reads(case_sample: CaseSample) -> bool:
        case: Case = case_sample.case
        sample: Sample = case_sample.sample
        if (
            case.data_analysis in [Workflow.MICROSALT, Workflow.TAXPROFILER]
            and not sample.is_negative_control
        ):
            return sample.reads >= sample.expected_reads_for_sample  # type: ignore Illumina sample
        else:
            return True



from cg.services.orders.validation.order_types.balsamic.models.sample import BalsamicSample
from cg.services.orders.validation.order_types.balsamic_umi.models.sample import BalsamicUmiSample
from cg.services.orders.validation.order_types.fastq.models.sample import FastqSample
from cg.services.orders.validation.order_types.fluffy.models.sample import FluffySample
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample
from cg.services.orders.validation.order_types.mip_rna.models.sample import MIPRNASample
from cg.services.orders.validation.order_types.raredisease.models.sample import RarediseaseSample
from cg.services.orders.validation.order_types.rml.models.sample import RMLSample
from cg.services.orders.validation.order_types.rna_fusion.models.sample import RNAFusionSample
from cg.services.orders.validation.order_types.tomte.models.sample import TomteSample

HumanSample = (
    BalsamicSample
    | BalsamicUmiSample
    | FastqSample
    | MIPDNASample
    | RarediseaseSample
    | RNAFusionSample
    | TomteSample
)

IndexedSample = FluffySample | RMLSample

SampleInCase = (
    BalsamicSample
    | BalsamicUmiSample
    | MIPDNASample
    | MIPRNASample
    | RarediseaseSample
    | RNAFusionSample
    | TomteSample
)

SampleWithRelatives = MIPDNASample | RarediseaseSample | TomteSample

SampleWithSkipRC = FastqSample | MIPDNASample | RarediseaseSample | TomteSample

SampleWithCaptureKit = BalsamicSample | BalsamicUmiSample | FastqSample



"""CLI commands to post-process a sequencing run."""

import logging

import rich_click as click

from cg.cli.post_process.utils import (
    UnprocessedRunInfo,
    get_post_processing_service_from_run_full_name,
    get_unprocessed_runs_info,
)
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.cli_options import DRY_RUN
from cg.models.cg_config import CGConfig
from cg.services.run_devices.abstract_classes import PostProcessingService

LOG = logging.getLogger(__name__)


@click.group(name="post-process", context_settings=CLICK_CONTEXT_SETTINGS)
def post_process_group():
    """Post-process sequencing runs from the sequencing instruments."""
    LOG.info("Running cg post-processing.")


@post_process_group.command(name="run")
@DRY_RUN
@click.argument("run-full-name")
@click.pass_obj
def post_process_run(context: CGConfig, run_full_name: str, dry_run: bool) -> None:
    """Post-process a sequencing run from the PacBio instrument.

    run-full-name is the full name of the sequencing unit of run. For example:
        PacBio: 'r84202_20240522_133539/1_A01'
    """
    post_processing_service: PostProcessingService = get_post_processing_service_from_run_full_name(
        context=context, run_full_name=run_full_name
    )
    post_processing_service.post_process(run_full_name=run_full_name, dry_run=dry_run)


@post_process_group.command(name="all")
@DRY_RUN
@click.option(
    "--instrument",
    type=click.Choice(["pacbio", "all"]),
    default="all",
    show_default=True,
    help="The instrument for which the runs will be post-processed. Choose 'all' to post-process runs from all instruments.",
    required=False,
)
@click.pass_obj
def post_process_all_runs(context: CGConfig, instrument: str, dry_run: bool) -> None:
    """Post-process all available runs."""
    exit_success: bool = True
    unprocessed_runs: list[UnprocessedRunInfo] = get_unprocessed_runs_info(
        context=context, instrument=instrument
    )
    for run in unprocessed_runs:
        try:
            run.post_processing_service.post_process(run_full_name=run.name, dry_run=dry_run)
        except Exception as error:
            LOG.error(f"Could not post-process {run.instrument} run {run.name}: {error}")
            exit_success = False
    if not exit_success:
        raise click.Abort


post_process_group.add_command(post_process_run)
post_process_group.add_command(post_process_all_runs)



import logging
from http import HTTPStatus
from flask import Blueprint, abort, g, jsonify

from cg.server.endpoints.utils import before_request

LOG = logging.getLogger(__name__)
USERS_BLUEPRINT = Blueprint("users", __name__, url_prefix="/api/v1")
USERS_BLUEPRINT.before_request(before_request)


@USERS_BLUEPRINT.route("/me")
def get_user_information():
    """Return information about current user."""
    if not g.current_user.is_admin and not g.current_user.customers:
        LOG.error(
            "%s is not admin and is not connected to any customers, aborting", g.current_user.email
        )
        return abort(HTTPStatus.FORBIDDEN)

    return jsonify(user=g.current_user.to_dict())



from cg.constants.constants import SexOptions
from cg.models.qc_metrics import QCMetrics


class RarediseaseQCMetrics(QCMetrics):
    """Raredisease QC metrics."""

    percent_duplication: float
    picard_pct_pf_reads_aligned: float
    predicted_sex_sex_check: SexOptions



"""Gene panel specific constants."""

from enum import StrEnum

from cg.constants.constants import CustomerId

GENOME_BUILD_37: str = "37"
GENOME_BUILD_38: str = "GRCh38"


class GenePanelGenomeBuild(StrEnum):
    hg19 = GENOME_BUILD_37
    hg38 = GENOME_BUILD_38


class GenePanelMasterList(StrEnum):
    BRAIN = "BRAIN"
    CARDIOLOGY = "Cardiology"
    CILM = "CILM"
    CH = "CH"
    CHD = "CHD"
    CHILDHYPOTONIA = "CHILDHYPOTONIA"
    CTD = "CTD"
    COMPD = "COMPD"
    DIAB = "DIAB"
    ENDO = "ENDO"
    EP = "EP"
    HEARING = "HEARING"
    HYDRO = "HYDRO"
    IBMFS = "IBMFS"
    IEM = "IEM"
    IF = "IF"
    MCARTA = "mcarta"
    MHT = "MHT"
    MIT = "MIT"
    MOVE = "MOVE"
    MTDNA = "mtDNA"
    NBS_M = "NBS-M"
    NEURODEG = "NEURODEG"
    NOONAN = "NOONAN"
    NMD = "NMD"
    OMIM_AUTO = "OMIM-AUTO"
    OPTIC = "OPTIC"
    PANELAPP_GREEN = "PANELAPP-GREEN"
    PEDHEP = "PEDHEP"
    PID = "PID"
    PIDCAD = "PIDCAD"
    SKD = "SKD"
    SOVM = "SOVM"
    STROKE = "STROKE"
    AID = "AID"
    IHC = "IHC"
    VEO_IBD = "VEO-IBD"
    KIDNEY = "Kidney"
    CAKUT = "CAKUT"
    NCRNA = "ncRNA"
    # Disorders of Sex Development
    DSD = "DSD"
    DSD_S = "DSD-S"
    POI = "POI"
    # Ophthalmology
    ALBINISM = "ALBINISM"
    ANTE_ED = "ANTE-ED"
    OPHTHALMO = "OPHTHALMO"
    RETINA = "RETINA"
    SED = "SED"

    @classmethod
    def get_panel_names(cls, panels=None) -> list[str]:
        """Return requested panel names from the Master list, or all panels if none are specified."""
        return list(panels) if panels else list(map(lambda panel: panel.value, cls))

    @staticmethod
    def collaborators() -> set[str]:
        """Return collaborators of the Master list."""
        return {
            CustomerId.CG_INTERNAL_CUSTOMER,
            CustomerId.CUST002,
            CustomerId.CUST003,
            CustomerId.CUST004,
            CustomerId.CUST042,
        }

    @staticmethod
    def is_customer_collaborator_and_panels_in_gene_panels_master_list(
        customer_id: str, gene_panels: set[str]
    ) -> bool:
        return customer_id in GenePanelMasterList.collaborators() and gene_panels.issubset(
            GenePanelMasterList.get_panel_names()
        )

    @staticmethod
    def get_non_specific_gene_panels() -> set[str]:
        return {GenePanelMasterList.OMIM_AUTO, GenePanelMasterList.PANELAPP_GREEN}



from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.raredisease.models.sample import RarediseaseSample


class RarediseaseCase(Case[RarediseaseSample]):
    cohorts: list[str] | None = None
    panels: list[str]
    synopsis: str | None = None

    def get_samples_with_father(self) -> list[tuple[RarediseaseSample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.father]

    def get_samples_with_mother(self) -> list[tuple[RarediseaseSample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.mother]



from typing import TypeVar

from pydantic_core import ValidationError

from cg.services.orders.validation.errors.validation_errors import ValidationErrors
from cg.services.orders.validation.model_validator.utils import convert_errors
from cg.services.orders.validation.models.order import Order

ParsedOrder = TypeVar("ParsedOrder", bound=Order)


class ModelValidator:

    @staticmethod
    def validate(order: dict, model: type[Order]) -> tuple[ParsedOrder | None, ValidationErrors]:
        try:
            order: Order = model.model_validate(order)
            return order, ValidationErrors()
        except ValidationError as error:
            return None, convert_errors(pydantic_errors=error)



import logging
import re

from housekeeper.store.models import File, Version

from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.constants.constants import FileFormat
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG, AnalysisTag, NFAnalysisTags
from cg.constants.scout import (
    RANK_MODEL_THRESHOLD,
    RAREDISEASE_CASE_TAGS,
    RAREDISEASE_SAMPLE_TAGS,
    GenomeBuild,
    UploadTrack,
)
from cg.constants.sequencing import Variants
from cg.io.controller import ReadFile
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.scout.scout_load_config import (
    CaseImages,
    CustomImages,
    Eklipse,
    RarediseaseLoadConfig,
    ScoutRarediseaseIndividual,
)
from cg.store.models import Analysis

LOG = logging.getLogger(__name__)


class RarediseaseConfigBuilder(ScoutConfigBuilder):
    def __init__(
        self,
        raredisease_analysis_api: RarediseaseAnalysisAPI,
        lims_api: LimsAPI,
        madeline_api: MadelineAPI,
    ):
        super().__init__(
            lims_api=lims_api,
        )
        self.case_tags: CaseTags = CaseTags(**RAREDISEASE_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**RAREDISEASE_SAMPLE_TAGS)
        self.raredisease_analysis_api: RarediseaseAnalysisAPI = raredisease_analysis_api
        self.lims_api: LimsAPI = lims_api
        self.madeline_api: MadelineAPI = madeline_api

    def build_load_config(self, hk_version: Version, analysis: Analysis) -> RarediseaseLoadConfig:
        """Create a RAREDISEASE specific load config for uploading analysis to Scout."""
        LOG.info("Build load config for RAREDISEASE case")
        load_config: RarediseaseLoadConfig = RarediseaseLoadConfig(
            track=UploadTrack.RARE_DISEASE.value,
            delivery_report=self.get_file_from_hk(
                hk_tags={HK_DELIVERY_REPORT_TAG}, hk_version=hk_version
            ),
        )
        self.add_common_info_to_load_config(load_config=load_config, analysis=analysis)
        load_config.gene_panels = self.raredisease_analysis_api.get_aggregated_panels(
            customer_id=analysis.case.customer.internal_id,
            default_panels=set(analysis.case.panels),
        )
        self.include_case_files(load_config=load_config, hk_version=hk_version)
        self.get_sample_information(
            load_config=load_config, analysis=analysis, hk_version=hk_version
        )
        self.include_pedigree_picture(load_config=load_config, analysis=analysis)
        self.load_custom_image_sample(
            load_config=load_config, analysis=analysis, hk_version=hk_version
        )
        load_config.human_genome_build = GenomeBuild.hg19
        load_config.rank_score_threshold = RANK_MODEL_THRESHOLD
        load_config.rank_model_version = self.get_rank_model_version(
            variant_type=Variants.SNV, hk_version=hk_version
        )
        load_config.sv_rank_model_version = self.get_rank_model_version(
            variant_type=Variants.SV, hk_version=hk_version
        )
        return load_config

    def get_rank_model_version(self, variant_type: Variants, hk_version: Version) -> str:
        """
        Returns the rank model version for a variant type from the manifest file.
        Raises:
            FileNotFoundError if no manifest file is found in housekeeper.
        """
        hk_manifest_file: File = self.get_file_from_hk(
            {NFAnalysisTags.MANIFEST}, hk_version=hk_version
        )
        if not hk_manifest_file:
            raise FileNotFoundError("No manifest file found in Housekeeper.")
        return self.extract_rank_model_from_manifest(
            hk_manifest_file=hk_manifest_file, variant_type=variant_type
        )

    def extract_rank_model_from_manifest(
        self, hk_manifest_file: File, variant_type: Variants
    ) -> str:
        content: dict[str, dict[str, str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=hk_manifest_file
        )
        return self.get_rank_model_version_from_manifest_content(
            content=content, variant_type=variant_type
        )

    def get_rank_model_version_from_manifest_content(
        self, content: dict[str, dict[str, str]], variant_type: Variants
    ) -> str:
        """
        Return the rank model version from the manifest file content.
        Raises:
            ValueError if pattern not found ing process or clinical not found in script.
        """
        pattern: str = variant_type.upper() + ":GENMOD_SCORE"
        for key, value in content["tasks"].items():
            process: str = value.get("process")
            script: str = value.get("script")
            if pattern in process and AnalysisTag.CLINICAL in script:
                return self._get_version_from_manifest_script(script)
        raise ValueError(
            f"Either {pattern} not found in any process or {AnalysisTag.CLINICAL} not found in any script of the manifest file"
        )

    def _get_version_from_manifest_script(self, script: str) -> str:
        """
        Returns the rank model version in the format 'vX.X from the given script string.
        Raises a ValueError if no rank model version is found in the script string.
        """
        if match := re.search(r"v(\d+\.\d+)", script):
            return match.group(1)
        raise ValueError("No rank model version found")

    def load_custom_image_sample(
        self, load_config: RarediseaseLoadConfig, analysis: Analysis, hk_version: Version
    ) -> None:
        """Build custom images config."""
        LOG.info("Adding custom images")

        eklipse_images: list = []
        for sample in analysis.case.samples:
            sample_id: str = sample.internal_id
            eklipse_image_path = self.get_file_from_hk(
                hk_tags=set(self.sample_tags.eklipse_path).union({sample_id}),
                hk_version=hk_version,
            )
            if eklipse_image_path:
                eklipse_image = Eklipse(
                    title=sample_id,
                    path=eklipse_image_path,
                    description="eKLIPse MT images",
                    width="800",
                    height="800",
                )
                eklipse_images.append(eklipse_image)
        if eklipse_images:
            case_images = CaseImages(eKLIPse=eklipse_images)
            config_custom_images = CustomImages(case_images=case_images)
            load_config.custom_images = config_custom_images

    def include_case_files(self, load_config: RarediseaseLoadConfig, hk_version: Version) -> None:
        """Include case level files for RAREDISEASE case."""
        LOG.info("Including RAREDISEASE specific case level files")
        for scout_key in RAREDISEASE_CASE_TAGS.keys():
            self._include_case_file(
                load_config=load_config, scout_key=scout_key, hk_version=hk_version
            )

    def _include_case_file(
        self, load_config: RarediseaseLoadConfig, scout_key: str, hk_version: Version
    ) -> None:
        """Include the file path associated to a scout configuration parameter if the corresponding housekeeper tags
        are found. Returns None."""
        file_path = self.get_file_from_hk(getattr(self.case_tags, scout_key), hk_version=hk_version)
        setattr(load_config, scout_key, file_path)

    def include_sample_files(
        self, config_sample: ScoutRarediseaseIndividual, hk_version: Version
    ) -> None:
        """Include sample level files that are optional for mip samples."""
        LOG.info("Including RAREDISEASE specific sample level files")
        sample_id: str = config_sample.sample_id
        config_sample.vcf2cytosure = self.get_sample_file(
            hk_tags=self.sample_tags.vcf2cytosure,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.mt_bam = self.get_sample_file(
            hk_tags=self.sample_tags.mt_bam, sample_id=sample_id, hk_version=hk_version
        )
        config_sample.chromograph_images.autozygous = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_autozyg,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.coverage = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_coverage,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.upd_regions = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_regions,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.upd_sites = self.remove_chromosome_substring(
            file_path=self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_sites,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.mitodel_file = self.get_sample_file(
            hk_tags=self.sample_tags.mitodel_file,
            sample_id=sample_id,
            hk_version=hk_version,
        )
        self.include_reviewer_files(config_sample=config_sample, hk_version=hk_version)



"""Module for the PacBioHousekeeperService used in the Post processing flow."""

import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FileExtensions
from cg.constants.housekeeper_tags import AlignmentFileTag
from cg.constants.pacbio import (
    PacBioBundleTypes,
    PacBioDirsAndFiles,
    PacBioHousekeeperTags,
    file_pattern_to_bundle_type,
)
from cg.services.run_devices.abstract_classes import PostProcessingHKService
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingParsingError,
    PostProcessingRunFileManagerError,
    PostProcessingStoreFileError,
)
from cg.services.run_devices.pacbio.housekeeper_service.models import PacBioFileData
from cg.services.run_devices.pacbio.metrics_parser.metrics_parser import PacBioMetricsParser
from cg.services.run_devices.pacbio.metrics_parser.models import PacBioMetrics
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager
from cg.utils.mapping import get_item_by_pattern_in_source

LOG = logging.getLogger(__name__)


class PacBioHousekeeperService(PostProcessingHKService):

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        file_manager: PacBioRunFileManager,
        metrics_parser: PacBioMetricsParser,
    ):
        self.hk_api: HousekeeperAPI = hk_api
        self.file_manager: PacBioRunFileManager = file_manager
        self.metrics_parser: PacBioMetricsParser = metrics_parser

    @handle_post_processing_errors(
        to_except=(PostProcessingRunFileManagerError, PostProcessingParsingError),
        to_raise=PostProcessingStoreFileError,
    )
    def store_files_in_housekeeper(self, run_data: PacBioRunData, dry_run: bool = False) -> None:
        parsed_metrics: PacBioMetrics = self.metrics_parser.parse_metrics(run_data)
        file_to_store: list[Path] = self.file_manager.get_files_to_store(run_data)
        for file_path in file_to_store:
            bundle_info: PacBioFileData = self._create_bundle_info(
                file_path=file_path, parsed_metrics=parsed_metrics
            )
            if dry_run:
                LOG.info(f"Dry run: would have added {bundle_info.file_path} to Housekeeper.")
                continue
            self.hk_api.create_bundle_and_add_file_with_tags(
                bundle_name=bundle_info.bundle_name,
                file_path=bundle_info.file_path,
                tags=bundle_info.tags,
            )
        LOG.debug(f"Files stored in Housekeeper for run {run_data.run_id}")

    @staticmethod
    def _get_bundle_type_for_file(file_path: Path) -> str:
        return get_item_by_pattern_in_source(
            source=file_path.name, pattern_map=file_pattern_to_bundle_type
        )

    @staticmethod
    def _get_sample_id_from_barcode(barcode: str, metrics: PacBioMetrics) -> str:
        full_barcode: str = f"{barcode}--{barcode}"
        for sample in metrics.samples:
            if sample.barcode_name == full_barcode:
                return sample.sample_internal_id
        raise PostProcessingStoreFileError(f"Sample not found for barcode: {barcode}")

    @staticmethod
    def _get_tags_for_file(file_path: Path) -> list[str]:
        file_pattern_to_tag: dict[str, list[str]] = {
            PacBioDirsAndFiles.BARCODES_REPORT: [PacBioHousekeeperTags.BARCODES_REPORT],
            PacBioDirsAndFiles.CONTROL_REPORT: [PacBioHousekeeperTags.CONTROL_REPORT],
            f".*{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}$": [PacBioHousekeeperTags.CCS_REPORT],
            PacBioDirsAndFiles.LOADING_REPORT: [PacBioHousekeeperTags.LOADING_REPORT],
            PacBioDirsAndFiles.RAW_DATA_REPORT: [PacBioHousekeeperTags.RAWDATA_REPORT],
            PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT: [PacBioHousekeeperTags.DATASETS_REPORT],
            f"{PacBioDirsAndFiles.HIFI_READS}.*{FileExtensions.BAM}$": [AlignmentFileTag.BAM],
            f".*{PacBioDirsAndFiles.METADATA_FILE_SUFFIX}$": [PacBioHousekeeperTags.METADATA],
        }
        return get_item_by_pattern_in_source(source=file_path.name, pattern_map=file_pattern_to_tag)

    def _create_bundle_info(self, file_path: Path, parsed_metrics: PacBioMetrics) -> PacBioFileData:
        tags: list[str] = self._get_tags_for_file(file_path)
        tags.append(parsed_metrics.dataset_metrics.cell_id)
        if self._is_file_type_smrt_cell(file_path):
            bundle_name: str = parsed_metrics.dataset_metrics.cell_id
        else:
            sample_internal_id: str = self._get_sample_internal_id_from_file(
                file_path=file_path, parsed_metrics=parsed_metrics
            )
            tags.append(sample_internal_id)
            bundle_name: str = sample_internal_id
        return PacBioFileData(
            bundle_name=bundle_name,
            file_path=file_path,
            tags=tags,
        )

    def _get_sample_internal_id_from_file(
        self, file_path: Path, parsed_metrics: PacBioMetrics
    ) -> str:
        barcode: str = file_path.name.split(".")[-2]
        return self._get_sample_id_from_barcode(barcode=barcode, metrics=parsed_metrics)

    def _is_file_type_smrt_cell(self, file_path: Path) -> bool:
        return self._get_bundle_type_for_file(file_path) == PacBioBundleTypes.SMRT_CELL



from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.order_types.balsamic_umi.models.sample import BalsamicUmiSample


class BalsamicUmiCase(Case[BalsamicUmiSample]):
    cohorts: list[str] | None = None
    synopsis: str | None = None



from pathlib import Path

from cg.services.run_devices.run_names.service import RunNamesService


class PacbioRunNamesService(RunNamesService):

    def get_run_full_names(self) -> list[str]:
        """
        Get all the SMRTcell full names from the PacBio sequencing directory in the form
        <sequencing_run>/<SMRTcell>, for example:
            r84202_20240913_121403/1_C01
        """

        run_names = []
        for run_folder in Path(self.run_directory).iterdir():
            if run_folder.is_dir():
                for cell_folder in run_folder.iterdir():
                    run_names.append(f"{run_folder.name}/{cell_folder.name}")
        return run_names



from pydantic import BeforeValidator, Field
from typing_extensions import Annotated

from cg.models.orders.sample_base import NAME_PATTERN, ControlEnum, SexEnum
from cg.services.orders.validation.constants import ElutionBuffer, TissueBlockEnum
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_buffer, parse_control


class RNAFusionSample(Sample):
    age_at_sampling: float | None = None
    concentration_ng_ul: float | None = None
    control: Annotated[ControlEnum, BeforeValidator(parse_control)] = ControlEnum.not_control
    elution_buffer: Annotated[ElutionBuffer | None, BeforeValidator(parse_buffer)] = None
    formalin_fixation_time: int | None = None
    phenotype_groups: list[str] | None = None
    phenotype_terms: list[str] | None = None
    post_formalin_fixation_time: int | None = None
    require_qc_ok: bool = False
    sex: SexEnum
    source: str
    source_comment: str | None = None
    subject_id: str = Field(pattern=NAME_PATTERN, min_length=1, max_length=128)
    tissue_block_size: TissueBlockEnum | None = None
    tumour: bool = True  # Always set to True upon submission, but if False, we show a warning



import logging

from cg.apps.lims import LimsAPI
from cg.constants import DataDelivery, Workflow
from cg.models.lims.sample import LimsSample
from cg.services.orders.validation.models.sample import Sample

LOG = logging.getLogger(__name__)


class OrderLimsService:

    def __init__(self, lims_api: LimsAPI):
        self.lims_api = lims_api

    @staticmethod
    def _build_lims_sample(
        customer: str,
        samples: list[Sample],
        workflow: Workflow,
        delivery_type: DataDelivery,
        skip_reception_control: bool,
    ) -> list[LimsSample]:
        """Convert order input to LIMS interface input."""
        samples_lims = []
        for sample in samples:
            dict_sample = sample.model_dump()
            LOG.debug(f"{sample.name}: prepare LIMS input")
            dict_sample["customer"] = customer
            dict_sample["data_analysis"] = workflow
            dict_sample["data_delivery"] = delivery_type
            dict_sample["family_name"] = sample._case_name
            if not dict_sample.get("priority"):
                dict_sample["priority"] = sample._case_priority
            if skip_reception_control:
                dict_sample["skip_reception_control"] = True
            if dict_sample.get("source") and dict_sample["source"] == "other":
                dict_sample["source"] = dict_sample.get("source_comment", "other")
            lims_sample: LimsSample = LimsSample.parse_obj(dict_sample)
            samples_lims.append(lims_sample)
        return samples_lims

    def process_lims(
        self,
        samples: list[Sample],
        customer: str,
        ticket: int | None,
        order_name: str,
        workflow: Workflow,
        delivery_type: DataDelivery,
        skip_reception_control: bool,
    ) -> tuple[any, dict]:
        """Process samples to add them to LIMS."""
        samples_lims: list[LimsSample] = self._build_lims_sample(
            customer=customer,
            samples=samples,
            workflow=workflow,
            delivery_type=delivery_type,
            skip_reception_control=skip_reception_control,
        )
        project_name: str = str(ticket) or order_name
        # Create new lims project
        project_data = self.lims_api.submit_project(
            project_name, [lims_sample.dict() for lims_sample in samples_lims]
        )
        lims_map: dict[str, str] = self.lims_api.get_samples(
            projectlimsid=project_data["id"], map_ids=True
        )
        return project_data, lims_map



from pathlib import Path
from typing import Any, Type
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from cg.constants.constants import FileFormat
from cg.constants.pacbio import MetricsFileFields, PacBioDirsAndFiles
from cg.exc import XMLError
from cg.io.controller import ReadFile
from cg.services.run_devices.pacbio.metrics_parser.models import (
    BarcodeMetrics,
    BaseMetrics,
    ControlMetrics,
    MetadataMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SampleMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.utils.files import get_file_with_pattern_from_list


def _get_data_model_from_pattern(pattern: str) -> Type[BaseMetrics]:
    """Return the data model based on the pattern."""
    pattern_to_model = {
        PacBioDirsAndFiles.BARCODES_REPORT: BarcodeMetrics,
        PacBioDirsAndFiles.CCS_REPORT_SUFFIX: ReadMetrics,
        PacBioDirsAndFiles.CONTROL_REPORT: ControlMetrics,
        PacBioDirsAndFiles.LOADING_REPORT: ProductivityMetrics,
        PacBioDirsAndFiles.RAW_DATA_REPORT: PolymeraseMetrics,
        PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT: SmrtlinkDatasetsMetrics,
    }
    return pattern_to_model.get(pattern)


def _parse_report_to_model(report_file: Path, data_model: Type[BaseMetrics]) -> BaseMetrics:
    """Parse the metrics report to a data model."""
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report_file)
    if data_model == SmrtlinkDatasetsMetrics:
        return data_model.model_validate(parsed_json[0])
    metrics: list[dict[str, Any]] = parsed_json.get(MetricsFileFields.ATTRIBUTES)
    data: dict = {
        report_field[MetricsFileFields.ID]: report_field[MetricsFileFields.VALUE]
        for report_field in metrics
    }
    return data_model.model_validate(data)


def get_parsed_metrics_from_file_name(metrics_files: list[Path], file_name: str) -> BaseMetrics:
    report_file: Path = get_file_with_pattern_from_list(files=metrics_files, pattern=file_name)
    data_model: Type[BaseMetrics] = _get_data_model_from_pattern(pattern=file_name)
    return _parse_report_to_model(report_file=report_file, data_model=data_model)


def _is_unassigned_reads_entry(sample_metric: SampleMetrics) -> bool:
    return sample_metric.sample_internal_id == "No Name"


def _parse_sample_data(sample_data: list[dict[str, Any]]) -> list[SampleMetrics]:
    """Parse all samples data into SampleMetrics given the sample section of the barcodes report."""
    sample_metrics: list[SampleMetrics] = []
    number_of_samples: int = len(sample_data[0].get(MetricsFileFields.VALUES))
    for sample_idx in range(number_of_samples):
        sample: dict = {}
        for data_field in sample_data:
            field_id: str = data_field.get(MetricsFileFields.ID)
            sample[field_id] = data_field.get(MetricsFileFields.VALUES)[sample_idx]
        sample_metric = SampleMetrics.model_validate(sample)
        if not _is_unassigned_reads_entry(sample_metric):
            sample_metrics.append(sample_metric)
    return sample_metrics


def get_parsed_sample_metrics(metrics_files: list[Path]) -> list[SampleMetrics]:
    """Parse the metrics from the barcodes report for each sample."""
    report_file: Path = get_file_with_pattern_from_list(
        files=metrics_files, pattern=PacBioDirsAndFiles.BARCODES_REPORT
    )
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report_file)
    sample_data: list[dict[str, Any]] = parsed_json.get(MetricsFileFields.TABLES)[0].get(
        MetricsFileFields.COLUMNS
    )
    return _parse_sample_data(sample_data)


def get_parsed_metadata_file(metrics_files: list[Path]) -> MetadataMetrics:
    namespaces: dict[str, str] = {
        "pb": "http://pacificbiosciences.com/PacBioDataModel.xsd"  # NOSONAR
    }
    metadata_file: Path = get_file_with_pattern_from_list(
        files=metrics_files, pattern=PacBioDirsAndFiles.METADATA_FILE_SUFFIX
    )
    root: Element = ElementTree.parse(metadata_file).getroot()
    run: Element | None = root.find(".//pb:Run", namespaces=namespaces)
    if run is not None:
        run_name, unique_id = run.get("Name"), run.get("UniqueId")
        if run_name and unique_id:
            return MetadataMetrics(run_name=run_name, unique_id=unique_id)
        else:
            raise XMLError(
                f"'Run' element is missing either 'Name' or 'UniqueId' in {metadata_file}"
            )
    else:
        raise XMLError(f"No 'Run' element found in {metadata_file}")



import logging
from pathlib import Path

from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import GenePanelMasterList
from cg.constants.gene_panel import GenePanelGenomeBuild
from cg.io.txt import write_txt_with_newlines
from cg.services.analysis_starter.configurator.file_creators.nextflow.utils import get_genome_build
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class GenePanelFileCreator:
    def __init__(self, store: Store, scout_api: ScoutAPI):
        self.store = store
        self.scout_api = scout_api

    def create(self, case_id: str, file_path: Path, double_hashtag_filtering: bool = False) -> None:
        content: list[str] = self._get_content(
            case_id=case_id, double_hashtag_filtering=double_hashtag_filtering
        )
        write_txt_with_newlines(file_path=file_path, content=content)
        LOG.info(f"Created gene panel file for case {case_id} at {file_path}")

    def _get_content(self, case_id: str, double_hashtag_filtering: bool) -> list[str]:
        case: Case = self.store.get_case_by_internal_id_strict(internal_id=case_id)
        genome_build: GenePanelGenomeBuild = get_genome_build(workflow=case.data_analysis)
        all_panels: list[str] = self._get_aggregated_panels(
            customer_id=case.customer.internal_id, default_panels=set(case.panels)
        )
        panels: list[str] = self.scout_api.export_panels(build=genome_build, panels=all_panels)
        if double_hashtag_filtering:
            panels = [panel for panel in panels if not panel.startswith("##")]
        return panels

    def _get_aggregated_panels(self, customer_id: str, default_panels: set[str]) -> list[str]:
        """Check if the customer is collaborator for gene panel master list
        and if all default panels are included in the gene panel master list.
        If not, add gene panel combo and broad non-specific gene panels.
        Return an aggregated gene panel."""
        if GenePanelMasterList.is_customer_collaborator_and_panels_in_gene_panels_master_list(
            customer_id=customer_id, gene_panels=default_panels
        ):
            return GenePanelMasterList.get_panel_names()
        all_panels: set[str] = default_panels | GenePanelMasterList.get_non_specific_gene_panels()
        return sorted(all_panels)



"""Module to group PDC related commands."""

import logging
from subprocess import CalledProcessError

import psutil

from cg.constants.pdc import DSMCParameters
from cg.constants.process import EXIT_WARNING
from cg.exc import PdcError
from cg.utils import Process

LOG = logging.getLogger(__name__)

SERVER = "hasta"
NO_FILE_FOUND_ANSWER = "ANS1092W"
MAX_NR_OF_DSMC_PROCESSES: int = 3


class PdcService:
    """Group PDC related commands"""

    def __init__(self, binary_path: str, dry_run: bool = False):
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    @classmethod
    def validate_is_dsmc_running(cls) -> bool:
        """Check if a Dsmc process is already running on the system.
        Raises:
            Exception: for all non-exit exceptions.
        """
        is_dsmc_running: bool = False
        dsmc_process_count: int = 0
        try:
            for process in psutil.process_iter():
                if "dsmc" == process.name():
                    dsmc_process_count += 1
        except Exception as error:
            LOG.debug(f"{error}")
        if dsmc_process_count >= MAX_NR_OF_DSMC_PROCESSES:
            is_dsmc_running = True
            LOG.debug("Too many Dsmc processes are already running")
        return is_dsmc_running

    def archive_file_to_pdc(self, file_path: str) -> None:
        """Archive a file by storing it on PDC."""
        if not self.dry_run:
            self.run_dsmc_command(command=DSMCParameters.ARCHIVE_COMMAND + [file_path])

    def query_pdc(self, search_pattern: str) -> None:
        """Query PDC based on a given search pattern."""
        command: list = DSMCParameters.QUERY_COMMAND.copy()
        command.append(search_pattern)
        self.run_dsmc_command(command=command)

    def retrieve_file_from_pdc(self, file_path: str, target_path: str = None) -> None:
        """Retrieve a file from PDC"""
        command: list = DSMCParameters.RETRIEVE_COMMAND.copy()
        command.append(file_path)
        if target_path:
            command.append(target_path)
        self.run_dsmc_command(command=command)

    def run_dsmc_command(self, command: list) -> None:
        """Runs a DSMC command.
        Raises:
            PdcError when unable to process command.
        """
        LOG.debug("Starting DSMC command:")
        LOG.debug(f"{self.process.binary} {' '.join(command)}")
        try:
            self.process.run_command(parameters=command, dry_run=self.dry_run)
        except CalledProcessError as error:
            if error.returncode == EXIT_WARNING:
                LOG.warning(f"{error}")
                return
            raise PdcError(message=f"{error}") from error

    @staticmethod
    def was_file_found(dsmc_output: str) -> bool:
        """Check if file was found in PDC."""
        return NO_FILE_FOUND_ANSWER not in dsmc_output



from pathlib import Path

from pydantic import BaseModel, field_validator


class FileDeliverable(BaseModel):
    """Specification for a general deliverables file."""

    id: str
    format: str
    path: str
    path_index: str | None = None
    step: str
    tag: str

    @field_validator("path", "path_index", mode="before")
    @classmethod
    def set_path_as_string(cls, file_path: str | Path) -> str | None:
        return str(Path(file_path)) if file_path else None


class WorkflowDeliverables(BaseModel):
    """Specification for workflow deliverables."""

    files: list[FileDeliverable]



from enum import IntEnum, StrEnum


class Slurm(StrEnum):
    PARTITION: str = "partition"


class SlurmProcessing(IntEnum):
    MAX_NODE_MEMORY: int = 180


SLURM_UPLOAD_MEMORY = 1
SLURM_UPLOAD_MAX_HOURS = 24
SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES = "--exclude=gpu-compute-0-[0-1],cg-dragen"
SLURM_UPLOAD_TASKS = 1



from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.order_types.balsamic.models.sample import BalsamicSample


class BalsamicCase(Case[BalsamicSample]):
    cohorts: list[str] | None = None
    synopsis: str | None = None



import rich_click as click

OPTION_PANEL_BED = click.option(
    "--panel-bed",
    required=False,
    help="Panel BED is determined based on capture kit \
    used for library prep. Set this option to override the default",
)
OPTION_WORKFLOW_PROFILE = click.option(
    "--workflow-profile",
    type=click.Path(exists=True),
    required=False,
    help="Path to directory containing config.yaml.",
)



from abc import abstractmethod
from pathlib import Path
from typing import Any

from pydantic import BaseModel

from cg.constants import SexOptions, Workflow
from cg.constants.constants import GenomeVersion
from cg.constants.priority import SlurmQos
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class BalsamicConfigInput(BaseModel):
    analysis_dir: Path
    analysis_workflow: Workflow
    artefact_snv_observations: Path
    balsamic_binary: Path
    balsamic_cache: Path
    cache_version: str
    cadd_annotations: Path
    cancer_germline_snv_observations: Path
    cancer_somatic_snv_observations: Path
    cancer_somatic_sv_observations: Path
    case_id: str
    clinical_snv_observations: Path
    clinical_sv_observations: Path
    conda_binary: Path
    conda_env: str
    fastq_path: Path
    gender: SexOptions
    genome_version: GenomeVersion
    gnomad_min_af5: Path
    normal_sample_name: str | None = None
    sentieon_install_dir: Path
    sentieon_license: str
    swegen_snv: Path
    swegen_sv: Path
    tumor_sample_name: str

    def dump_to_cli(self) -> str:
        """Dump the Balsamic case config to a CLI command. None flags are excluded and boolean flags are converted to
        only add the flag."""
        command = (
            f"{self.conda_binary} run --name {self.conda_env} {self.balsamic_binary} config case"
        )

        for flag, value in self._get_flags().items():
            if isinstance(value, bool):
                if value is True:
                    command += f" {flag}"
            elif value is not None:
                command += f" {flag} {value}"
        return command

    @abstractmethod
    def _get_flags(self) -> dict[str, Any]:
        pass


class BalsamicConfigInputPanel(BalsamicConfigInput):
    cancer_somatic_snv_panel_observations: Path | None = None
    exome: bool
    panel_bed: Path
    pon_cnn: Path | None = (
        None  # Equivalent to --gens-coverage-pon in wgs analysis, depends on the panel
    )
    soft_filter_normal: bool = False  # True for all panel analyses with a normal sample

    def _get_flags(self) -> dict[str, Any]:
        return {
            "--analysis-dir": self.analysis_dir,
            "--analysis-workflow": self.analysis_workflow,
            "--artefact-snv-observations": self.artefact_snv_observations,
            "--balsamic-cache": self.balsamic_cache,
            "--cache-version": self.cache_version,
            "--cadd-annotations": self.cadd_annotations,
            "--cancer-germline-snv-observations": self.cancer_germline_snv_observations,
            "--cancer-somatic-snv-observations": self.cancer_somatic_snv_observations,
            "--cancer-somatic-snv-panel-observations": self.cancer_somatic_snv_panel_observations,
            "--cancer-somatic-sv-observations": self.cancer_somatic_sv_observations,
            "--case-id": self.case_id,
            "--clinical-snv-observations": self.clinical_snv_observations,
            "--clinical-sv-observations": self.clinical_sv_observations,
            "--fastq-path": self.fastq_path,
            "--gender": self.gender,
            "--genome-version": self.genome_version,
            "--gnomad-min-af5": self.gnomad_min_af5,
            "--normal-sample-name": self.normal_sample_name,
            "--panel-bed": self.panel_bed,
            "--exome": self.exome,  # MUST be after panel bed
            "--pon-cnn": self.pon_cnn,
            "--sentieon-install-dir": self.sentieon_install_dir,
            "--sentieon-license": self.sentieon_license,
            "--soft-filter-normal": self.soft_filter_normal,
            "--swegen-snv": self.swegen_snv,
            "--swegen-sv": self.swegen_sv,
            "--tumor-sample-name": self.tumor_sample_name,
        }


class BalsamicConfigInputWGS(BalsamicConfigInput):
    artefact_sv_observations: Path
    genome_interval: Path
    gens_coverage_pon: Path  # Equivalent to --pon-cnn in panel analysis, depends on the sex

    def _get_flags(self) -> dict[str, Any]:
        return {
            "--analysis-dir": self.analysis_dir,
            "--analysis-workflow": self.analysis_workflow,
            "--artefact-snv-observations": self.artefact_snv_observations,
            "--artefact-sv-observations": self.artefact_sv_observations,
            "--balsamic-cache": self.balsamic_cache,
            "--cache-version": self.cache_version,
            "--cadd-annotations": self.cadd_annotations,
            "--cancer-germline-snv-observations": self.cancer_germline_snv_observations,
            "--cancer-somatic-snv-observations": self.cancer_somatic_snv_observations,
            "--cancer-somatic-sv-observations": self.cancer_somatic_sv_observations,
            "--case-id": self.case_id,
            "--clinical-snv-observations": self.clinical_snv_observations,
            "--clinical-sv-observations": self.clinical_sv_observations,
            "--fastq-path": self.fastq_path,
            "--gender": self.gender,
            "--genome-interval": self.genome_interval,
            "--genome-version": self.genome_version,
            "--gens-coverage-pon": self.gens_coverage_pon,
            "--gnomad-min-af5": self.gnomad_min_af5,
            "--normal-sample-name": self.normal_sample_name,
            "--sentieon-install-dir": self.sentieon_install_dir,
            "--sentieon-license": self.sentieon_license,
            "--swegen-snv": self.swegen_snv,
            "--swegen-sv": self.swegen_sv,
            "--tumor-sample-name": self.tumor_sample_name,
        }


class BalsamicCaseConfig(CaseConfig):
    account: str
    binary: Path
    conda_binary: Path
    environment: str
    head_job_partition: str
    qos: SlurmQos
    sample_config: Path
    workflow: Workflow
    workflow_profile: Path | None = None

    def get_start_command(self) -> str:
        command = (
            "{conda_binary} run --name {environment} {binary} run analysis --account {account} "
            "--qos {qos} --sample-config {sample_config} --headjob-partition {head_job_partition} "
            "--run-analysis".format(**self.model_dump())
        )
        if self.workflow_profile:
            command += f" --workflow-profile {self.workflow_profile}"
        return command



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.mutant.constants import MutantDeliveryType
from cg.services.orders.validation.order_types.mutant.models.sample import MutantSample


class MutantOrder(OrderWithSamples[MutantSample]):
    delivery_type: MutantDeliveryType



from pydantic import BeforeValidator
from typing_extensions import Annotated

from cg.models.orders.sample_base import ControlEnum, PriorityEnum
from cg.services.orders.validation.constants import ElutionBuffer
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_buffer, parse_control


class TaxprofilerSample(Sample):
    concentration_ng_ul: float | None = None
    control: Annotated[ControlEnum, BeforeValidator(parse_control)] = ControlEnum.not_control
    elution_buffer: Annotated[ElutionBuffer, BeforeValidator(parse_buffer)]
    priority: PriorityEnum
    quantity: int | None = None
    require_qc_ok: bool = False
    source: str
    source_comment: str | None = None



import logging
from pathlib import Path

from cg.constants.constants import ReadDirection
from cg.services.fastq_concatenation_service.utils import (
    remove_raw_fastqs,
    concatenate_fastq_reads_for_direction,
)

LOG = logging.getLogger(__name__)


class FastqConcatenationService:
    """Fastq file concatenation service."""

    @staticmethod
    def concatenate(
        sample_id: str,
        fastq_directory: Path,
        forward_output_path: Path,
        reverse_output_path: Path,
        remove_raw: bool = False,
    ):
        """Concatenate fastq files for a given sample in a directory and write the concatenated files to the output path.

        Args:
            sample_id: The identifier to identify the samples by it should be a unique identifier in the file name.
            fastq_directory: The directory containing the fastq files.
            forward_output_path: The path where the concatenated forward reads will be written.
            reverse_output_path: The path where the concatenated reverse reads will be written.
            remove_raw: If True, remove the raw fastq files after concatenation.
        """
        LOG.debug(
            f"[Concatenation Service] Concatenating fastq files for {sample_id} in {fastq_directory}"
        )
        temp_forward: Path | None = concatenate_fastq_reads_for_direction(
            directory=fastq_directory, sample_id=sample_id, direction=ReadDirection.FORWARD
        )
        temp_reverse: Path | None = concatenate_fastq_reads_for_direction(
            directory=fastq_directory, sample_id=sample_id, direction=ReadDirection.REVERSE
        )

        if remove_raw:
            remove_raw_fastqs(
                sample_id=sample_id,
                fastq_directory=fastq_directory,
                forward_file=temp_forward,
                reverse_file=temp_reverse,
            )

        if temp_forward:
            LOG.info(f"Concatenated forward reads to {forward_output_path}")
            temp_forward.rename(forward_output_path)

        if temp_reverse:
            LOG.info(f"Concatenated reverse reads to {reverse_output_path}")
            temp_reverse.rename(reverse_output_path)



"""CLI options for Nextflow and NF-Tower."""

import logging

import rich_click as click

from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.cli.workflow.utils import validate_force_store_option
from cg.constants.cli_options import COMMENT, DRY_RUN, FORCE
from cg.constants.constants import MetaApis
from cg.exc import CgError, HousekeeperStoreError
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


OPTION_RESUME = click.option(
    "--resume",
    default=True,
    show_default=True,
    help="Execute the script using the cached results, useful to continue "
    "executions that were stopped by an error",
)
OPTION_REVISION = click.option(
    "--revision",
    type=str,
    help="Revision of workflow to run (either a git branch, tag or commit SHA number)",
)


@click.command("metrics-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def metrics_deliver(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """
    Create and validate a metrics deliverables file for given case id.
    If QC metrics are met it sets the status in Trailblazer to complete.
    If failed, it sets it as failed and adds a comment with information of the failed metrics.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.metrics_deliver(case_id=case_id, dry_run=dry_run)
    except CgError as error:
        raise click.Abort() from error


@click.command("report-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def report_deliver(context: CGConfig, case_id: str, dry_run: bool, force: bool) -> None:
    """
    Create a Housekeeper deliverables file for a given case ID.

    Raises:
        click.Abort: If an error occurs during the deliverables report generation or validation.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.report_deliver(case_id=case_id, dry_run=dry_run, force=force)
    except CgError as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()


@click.command("store-housekeeper")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def store_housekeeper(context: CGConfig, case_id: str, dry_run: bool, force: bool) -> None:
    """
    Store a finished NF-analysis in Housekeeper.

    Raises:
        click.Abort: If an error occurs while storing a case bundle in Housekeeper.
    """
    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.store_analysis_housekeeper(case_id=case_id, dry_run=dry_run, force=force)
    except HousekeeperStoreError as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}!")
        raise click.Abort()


@click.command("store")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_context
def store(
    context: click.Context, case_id: str, comment: str | None, dry_run: bool, force: bool
) -> None:
    """
    Store deliverable files in Housekeeper after meeting QC metrics criteria.

    Raises:
        click.Abort: If an error occurs during the deliverables file generation, metrics
        validation, or storage processes.
    """
    validate_force_store_option(force=force, comment=comment)
    analysis_api: NfAnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.store(case_id=case_id, comment=comment, dry_run=dry_run, force=force)
    except Exception as error:
        LOG.error(repr(error))
        raise click.Abort()


@click.command("store-available")
@DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """
    Store finished analyses for cases marked as running in StatusDB and completed or QC in Trailblazer.

    Raises:
        click.Abort: If any error occurs during the storage process.
    """

    analysis_api: NfAnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]
    was_successful: bool = True
    for case in analysis_api.get_cases_to_store():
        LOG.info(f"Storing deliverables for {case.internal_id}")
        try:
            analysis_api.store(case_id=case.internal_id, dry_run=dry_run)
        except Exception as error:
            LOG.error(f"Error storing {case.internal_id}: {repr(error)}")
            was_successful = False
    if not was_successful:
        raise click.Abort()



"""CLI support to start microSALT."""

import logging
from pathlib import Path
from typing import cast

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, resolve_compression, store, store_available
from cg.constants.constants import Workflow
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.analysis_starter import AnalysisStarter
from cg.services.analysis_starter.configurator.implementations.microsalt import (
    MicrosaltConfigurator,
)
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory

LOG = logging.getLogger(__name__)

ARGUMENT_UNIQUE_IDENTIFIER = click.argument("unique_id", required=True, type=click.STRING)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def microsalt(context: click.Context) -> None:
    """Microbial workflow"""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis["analysis_api"] = MicrosaltAnalysisAPI(config=context.obj)


microsalt.add_command(store)
microsalt.add_command(store_available)
microsalt.add_command(resolve_compression)


@microsalt.command("config-case")
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(cg_config: CGConfig, case_id: str) -> None:
    """
    Configure a microSALT case so that it is ready to be run.
    """
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(MicrosaltConfigurator, factory.get_configurator(Workflow.MICROSALT))
    configurator.configure(case_id=case_id)


@microsalt.command("run")
@click.option(
    "-c",
    "--config",
    "config_file_path",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="optionally change the config file path",
)
@ARGUMENT_CASE_ID
@click.pass_obj
def run(
    cg_config: CGConfig,
    config_file_path: click.Path,
    case_id: str,
) -> None:
    """Run a preconfigured microSALT case."""
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.MICROSALT
    )
    analysis_starter.run(case_id=case_id, config_file=config_file_path)


@microsalt.command("start")
@ARGUMENT_CASE_ID
@click.pass_obj
def start(cg_config: CGConfig, case_id: str) -> None:
    """
    Generates config file, links fastq files and runs the microSALT analysis for the provided case.
    """
    LOG.info(f"Starting microSALT workflow for {case_id}")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.MICROSALT
    )
    analysis_starter.start(case_id)


@microsalt.command("start-available")
@click.pass_obj
def start_available(cg_config: CGConfig) -> None:
    """Starts all available microSALT cases."""
    LOG.info("Starting microSALT workflow for all available cases.")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.MICROSALT
    )
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort


@microsalt.command("qc")
@ARGUMENT_UNIQUE_IDENTIFIER
@click.pass_context
def qc_microsalt(context: click.Context, unique_id: str) -> None:
    """Perform QC on a microSALT case."""
    analysis_api: MicrosaltAnalysisAPI = context.obj.meta_apis["analysis_api"]
    metrics_file_path: Path = analysis_api.get_metrics_file_path(unique_id)
    try:
        LOG.info(f"Performing QC on case {unique_id}")
        analysis_api.quality_checker.quality_control(metrics_file_path)
    except IndexError:
        LOG.error(f"No existing analysis directories found for case {unique_id}.")



"""Chanjo1 coverage utilities for workflow analysis APIs."""

import logging

from cg.apps.coverage import ChanjoAPI
from cg.clients.chanjo2.models import CoverageMetricsChanjo1

LOG = logging.getLogger(__name__)


def chanjo1_get_sample_coverage(
    chanjo_api: ChanjoAPI, sample_id: str, gene_ids: list[int]
) -> CoverageMetricsChanjo1 | None:
    sample_coverage: dict = chanjo_api.sample_coverage(sample_id=sample_id, panel_genes=gene_ids)
    if sample_coverage:
        return CoverageMetricsChanjo1(
            coverage_completeness_percent=sample_coverage.get("mean_completeness"),
            mean_coverage=sample_coverage.get("mean_coverage"),
        )
    LOG.warning(f"Could not calculate sample coverage for: {sample_id}")
    return None
